import pandas as pd

def fast_merge(left, right, left_on, right_on, how='inner'):
	"""
	:type left: pd.DataFrame
	:type right: pd.DataFrame
	:type left_on: list[str]
	:type right_on: list[str]
	:type how: str
	:rtype: pd.DataFrame
	"""

	result = left.set_index(keys=left_on, drop=True).join(
		other=right.set_index(keys=right_on, drop=True),
		how=how, lsuffix='_x', rsuffix='_y'
	)
	result[left_on] = result.index
	result[right_on] = result.index
	result.reset_index(drop=True, inplace=True)

	return result


def get_month_name(number, abr=False):
	months_map = {
		1: 'January', 2: 'February', 3: 'March',
		4: 'April', 5: 'May', 6: 'June',
		7: 'July', 8: 'August', 9: 'September',
		10: 'October', 11: 'November', 12: 'December'
	}
	result = months_map[number]
	if abr:
		return result[:3]
	else:
		return result


def get_month_num(name):
	months_map = {
		'jan': 1, 'feb': 2, 'mar': 3, 'apr': 4, 'may': 5, 'jun': 6,
		'jul': 7, 'aug': 8, 'sep': 9, 'oct': 10, 'nov': 11, 'dec': 12
	}
	return months_map[str(name).lower()[:3]]


def get_months_list(abr=False):
	return [get_month_name(abr=abr, number=i) for i in range(1, 13)]



import base64
import base32hex
import hashlib
from ..collections import make_hashable


def hash_object(obj, base=64):
	"""
	:type obj: Obj
	:type base: int
	:rtype str
	"""
	if hasattr(obj, '__hash64__') and base == 64:
		return obj.__hash64__()
	elif hasattr(obj, '__hash32__') and base == 32:
		return obj.__hash32__()

	hash_maker = hashlib.sha256()
	hash_maker.update(repr(make_hashable(obj)).encode())
	if base == 64:
		return base64.b64encode(hash_maker.digest()).decode()
	elif base == 32:
		return base32hex.b32encode(hash_maker.digest()).replace('=', '-').replace('----', 'x032')
	else:
		raise ValueError(f'base{base} is unknown!')


def hash_file(path, base=64, block_size=65536):
	hash_maker = hashlib.sha256()
	with open(path, 'rb') as file:
		block = file.read(block_size)
		while len(block) > 0:
			hash_maker.update(block)
			block = file.read(block_size)
	if base == 64:
		return base64.b64encode(hash_maker.digest()).decode()
	elif base == 32:
		return base32hex.b32encode(hash_maker.digest()).replace('=', '-')
	else:
		raise ValueError(f'base{base} is unknown!')


from pandas import DataFrame as PandasDF
from pyspark.sql import DataFrame as SparkDF
from pyspark.sql import functions as f

from ...text import remove_non_alphanumeric as _remove_alpha_func
from ...text import convert_camel_to_snake


def standardize_name(
		x, camel_to_snake=True, remove_non_alphanumeric=True,
		replace_with = '_', join_by = '__', ignore_errors = False
):
	"""
	:type x: list[str] or tuple or str
	:type camel_to_snake: bool
	:type remove_non_alphanumeric: bool
	:type replace_with: str
	:type join_by: str
	:type ignore_errors: bool
	:rtype: str
	"""
	if isinstance(x, tuple):
		x = list(x)

	# remove empty strings and remove non alphanumeric
	if isinstance(x, list):
		x = [str(s).strip() for s in x if s != '']
		x = join_by.join(x)

	try:
		if remove_non_alphanumeric:
			x = _remove_alpha_func(s=x, replace_with=' ')
			x = x.strip()
			x = _remove_alpha_func(s=x, replace_with=replace_with)
		if camel_to_snake:
			x = convert_camel_to_snake(x, remove_non_alphanumeric=remove_non_alphanumeric)
		x = x.lower().replace(' ', '_')
	except Exception as e:
		if ignore_errors:
			print('Error was ignored for: "', x, '" ', e, sep='')
		else:
			raise e

	return x


def standardize_columns(data, inplace=False, camel_to_snake=True, remove_non_alphanumeric=True):
	"""
	:type data: PandasDF or SparkDF
	:type inplace: bool
	:type camel_to_snake: bool
	:rtype: PandasDF or SparkDF
	"""
	if isinstance(data, PandasDF):
		if inplace:
			new_data = data
		else:
			new_data = data.copy()

		new_data.columns = list(map(
			lambda x: standardize_name(
				x=x, camel_to_snake=camel_to_snake, remove_non_alphanumeric=remove_non_alphanumeric
			),
			list(new_data.columns)
		))

		return new_data

	elif isinstance(data, SparkDF):
		return data.select([
			f.col(col).alias(
				standardize_name(x=col, camel_to_snake=camel_to_snake, remove_non_alphanumeric=remove_non_alphanumeric)
			)
			for col in data.columns
		])



class Sampler:
	def __init__(self, sample_size = None, sample_ratio = None, num_splits = 1):
		self.sample_size = sample_size
		self.sample_ratio = sample_ratio
		self.num_splits = num_splits


class Splitter(Sampler):
	def __init__(self, sample_size = None, sample_ratio = 0.2):
		pass



from ...collections import OrderedSet


def get_rows(data):
	try:
		return data.rows
	except AttributeError:
		return OrderedSet(range(data.shape[0]))


def get_columns(data):
	columns = data.columns
	if isinstance(columns, OrderedSet):
		return columns
	else:
		return OrderedSet(columns)



from func_timeout import func_timeout, FunctionTimedOut

from pandas import DataFrame as PandasDF, Series as PandasSeries, concat
from sklearn.cluster import KMeans as SklearnKMeans
from sklearn.metrics import silhouette_score

from pyspark.sql import functions as f
from pyspark.sql.types import DoubleType, StringType
from pyspark.sql import DataFrame as SparkDF, Column as SparkColumn
from pyspark.sql.window import Window
from pyspark.ml.clustering import KMeans as SparkKMeans
from pyspark.ml.feature import VectorAssembler
from pyspark.ml.evaluation import ClusteringEvaluator
from pyspark.ml.linalg import Vectors

from scipy.spatial.distance import cdist
import numpy as np


def get_distortion(X, kmeans_model):
	return sum(np.min(cdist(X, kmeans_model.cluster_centers_, 'euclidean'), axis=1)) / X.shape[0]


class TimedOutKMeans:
	pass


_KMEANS_STATE = [
	'_n_clusters', '_n_jobs', '_kwargs', '_model_dict', '_env', '_features', '_evaluator', '_raise_timeout',
	'_timeout', '_timed_out'
]


class KMeans:
	def __init__(
			self, n_clusters, n_jobs=-1, timeout=None, evaluator=None, raise_timeout=True, **kwargs
	):
		"""
		:type n_clusters: int
		:type n_jobs: int
		:type env: str
		:type features: list[str]
		:type X: SparkDF or PandasDF
		:type raise_timeout: bool
		:type evaluator: ClusteringEvaluator
		"""
		self._n_clusters = n_clusters
		self._n_jobs = n_jobs
		self._kwargs = kwargs

		self._model_dict = None
		self._env = None
		self._features = None

		self._evaluator = evaluator
		self._raise_timeout = raise_timeout

		self._timeout = timeout
		self._timed_out = False

	def __getstate__(self):
		return {
			name: getattr(self, name)
			for name in _KMEANS_STATE
		}

	def __setstate__(self, state):
		for name, value in state.items():
			setattr(self, name, value)

	#  PYTHON #

	def _fit_python(self, X):
		"""
		:type X: PandasDF
		"""
		self._features = list(X.columns)
		model = SklearnKMeans(
			n_clusters=self.n_clusters, n_jobs=self._n_jobs, **self._kwargs
		)
		if self._timeout:
			try:
				func_timeout(timeout=self._timeout, func=model.fit, kwargs={'X': X})
				distortion = get_distortion(X=X, kmeans_model=model)
				inertia = model.inertia_
				predictions = model.predict(X=X)
				score = silhouette_score(X, predictions, metric='euclidean')

			except FunctionTimedOut as e:
				self._timed_out = True
				if self._raise_timeout:
					raise e
				else:
					model = TimedOutKMeans()
				distortion = None
				inertia = None
				score = None
		else:
			model.fit(X=X)
			distortion = get_distortion(X=X, kmeans_model=model)
			inertia = model.inertia_
			predictions = model.predict(X=X)
			score = silhouette_score(X, predictions, metric='euclidean')

		self._model_dict = {
			'kmeans': model,
			'distortion': distortion,
			'inertia': inertia,
			'silhouette_score': score
		}
		return self

	@property
	def _python_model(self):
		"""
		:rtype: SklearnKMeans
		"""
		return self._model_dict['kmeans']

	@property
	def _python_cluster_centers(self):
		"""
		:rtype: list[np_array]
		"""
		return list(self._python_model.cluster_centers_)

	def _transform_python(self, X, prefix='distance_', keep_columns=False):
		"""
		:type X: PandasDF
		:type prefix: str
		:rtype: PandasDF
		"""
		result = PandasDF(
			self._python_model.transform(X=X[self._features]),
			columns=[f'{prefix}{x + 1}' for x in range(self._n_clusters)]
		)
		if isinstance(X, PandasDF):
			result.index = X.index

		if not keep_columns:
			return result

		elif isinstance(keep_columns, (list, str)):
			if isinstance(keep_columns, str):
				keep_columns = [keep_columns]
		else:
			keep_columns = list(X.columns)

		return concat([X[keep_columns], result], axis=1)

	def _predict_python(self, X, prefix='cluster_', keep_columns=False, return_dataframe=False):
		"""
		:type X: PandasDF
		:type prefix: str
		:type keep_columns: bool or str or list[str]
		:type return_dataframe: bool
		:rtype: PandasSeries or PandasDF
		"""
		ndarray = self._python_model.predict(X=X[self._features])
		series = PandasSeries(ndarray, name='prediction')
		predictions = series.apply(lambda x: f'{prefix}{x + 1}')
		if keep_columns:
			if isinstance(keep_columns, str):
				keep_columns = [keep_columns]

			result = X.copy()
			if isinstance(keep_columns, list):
				result = result[keep_columns]

			result['prediction'] = predictions
			return result

		elif return_dataframe:
			return predictions.to_frame(name='prediction')

		else:
			return predictions

	# SPARK #

	def _fit_spark(self, X):
		"""
		:type X: SparkDF
		"""
		self._features = X.columns
		assembler = VectorAssembler(inputCols=self._features, outputCol="features")
		assembled = assembler.transform(X)
		kmeans = SparkKMeans(k=self.n_clusters).fit(assembled.select('features'))
		if self._evaluator is not None:
			clustered = kmeans.transform(assembled)
			score = self._evaluator.evaluate(clustered)
			inertia = kmeans.summary.trainingCost

		else:
			score = None
			inertia = None

		self._model_dict = {
			'assembler': assembler,
			'kmeans': kmeans,
			'silhouette_score': score,
			'inertia': inertia
		}
		return self

	@property
	def _spark_model(self):
		"""
		:rtype: SparkKMeans
		"""
		return self._model_dict['kmeans']

	@property
	def _spark_assembler(self):
		"""
		:rtype: VectorAssembler
		"""
		return self._model_dict['assembler']

	@property
	def _spark_cluster_centers(self):
		"""
		:rtype: list[np_array]
		"""
		return self._spark_model.clusterCenters()

	def _transform_spark(self, X, prefix='distance_', keep_columns=False):
		"""
		:type X: SparkDF
		:type prefix: str
		:rtype: SparkDF
		"""
		assembled = self._spark_assembler.transform(X)
		centers = self.cluster_centers

		@f.udf(DoubleType())
		def _get_distance(features, cluster):
			"""
			:type features: np_array
			:rtype: SparkColumn
			"""
			center = centers[cluster]
			return float(Vectors.squared_distance(features, center) ** 0.5)

		if isinstance(keep_columns, (list, str)):
			if isinstance(keep_columns, str):
				keep_columns = [keep_columns]
		elif keep_columns:
			keep_columns = [col for col in X.columns if col != 'features']
		else:
			keep_columns = []

		transformed = assembled.select(*keep_columns, *[
			_get_distance('features', f.lit(cluster)).alias(f'{prefix}{cluster + 1}')
			for cluster in range(len(centers))
		])
		return transformed

	def _predict_spark(self, X, prefix='cluster_', keep_columns=False):
		"""
		:type X: SparkDF
		:rtype: SparkDF
		"""
		assembled = self._spark_assembler.transform(X)
		result = self._spark_model.transform(assembled)

		@f.udf(StringType())
		def _convert_prediction_to_string(x):
			return f'{prefix}{x + 1}'
		result = result.withColumn('prediction', _convert_prediction_to_string('prediction'))

		if keep_columns:
			if isinstance(keep_columns, str):
				keep_columns = [keep_columns]

			if isinstance(keep_columns, list):
				columns = keep_columns + ['prediction']
				return result.select(*columns)
			else:
				return result.drop('features')
		else:
			return result.select('prediction')

	# GENERAL METHODS #

	def fit(self, X):
		"""
		:type X: SparkDF or PandasDF
		"""
		if isinstance(X, SparkDF):
			self._env = 'spark'
		elif isinstance(X, PandasDF):
			self._env = 'python'
		else:
			raise RuntimeError(f'unknown data type: {type(X)}')

		if self._evaluator is None:
			self._evaluator = ClusteringEvaluator()

		if self._env == 'spark':
			return self._fit_spark(X=X)

		elif self._env == 'python':
			return self._fit_python(X=X)

		else:
			raise RuntimeError(f'unknown environment {self._env}')

	@property
	def cluster_centers(self):
		"""
		:rtype: list[np_array]
		"""
		if self._env == 'spark':
			return self._spark_cluster_centers
		else:
			return self._python_cluster_centers

	def transform(self, X, keep_columns=False):
		"""
		:type X: SparkDF or PandasDF
		:type keep_columns: bool
		"""
		if self._env == 'spark':
			return self._transform_spark(X=X, keep_columns=keep_columns)

		elif self._env == 'python':
			return self._transform_python(X=X, keep_columns=keep_columns)

		else:
			raise RuntimeError(f'unknown environment {self._env}')

	def fit_transform(self, X, keep_columns=False):
		"""
		:type X: SparkDF or PandasDF
		:type keep_columns: bool
		"""
		self.fit(X=X)
		return self.transform(X=X, keep_columns=keep_columns)

	def predict(self, X, prefix='cluster_', keep_columns=False, return_dataframe=None):
		"""
		:type X: SparkDF or PandasDF or PandasSeries
		:type prefix: str
		:type keep_columns: bool or str or list[str]
		:type return_dataframe: bool
		"""
		if self._env == 'spark':
			return self._predict_spark(X=X, prefix=prefix, keep_columns=keep_columns)

		elif self._env == 'python':
			return self._predict_python(
				X=X, prefix=prefix, keep_columns=keep_columns, return_dataframe=return_dataframe
			)

		else:
			raise RuntimeError(f'unknown environment {self._env}')

	def fit_predict(self, X, prefix='cluster_', keep_columns=False):
		self.fit(X=X)
		predictions = self.predict(X=X, prefix=prefix, keep_columns=keep_columns)
		return predictions

	def transform_and_predict(self, X, keep_columns=False):
		"""
		:type X: PandasDF or SparkDF
		:type keep_columns: bool or str or list[str]
		:rtype: PandasDF
		"""
		transformed = self.transform(X=X, keep_columns=keep_columns)
		predictions = self.predict(X=X, return_dataframe=False, keep_columns=False)

		if self._env == 'python':
			transformed['prediction'] = predictions

		elif self._env == 'spark':
			w = Window.orderBy(f.monotonically_increasing_id())
			transformed = transformed.withColumn('_index', f.row_number().over(w))
			predictions = predictions.withColumn('_index', f.row_number().over(w))
			transformed = transformed.join(predictions, on='_index', how='inner').drop('_index')

		else:
			raise RuntimeError(f'unknown environment {self._env}')

		return transformed

	@property
	def n_clusters(self):
		"""
		:rtype: int
		"""
		return self._n_clusters

	@property
	def timedout(self):
		return self._timed_out

	@property
	def distortion(self):
		return self._model_dict.get('distortion', None)

	@property
	def inertia(self):
		return self._model_dict.get('inertia', None)

	@property
	def silhouette_score(self):
		return self._model_dict.get('silhouette_score', None)



from itertools import chain, combinations


def get_powerset(iterable):
	"powerset([1,2,3]) --> [[], [1], [2], [3], [1, 2], [1, 3], [2, 3], [1, 2, 3]]"
	s = list(iterable)
	return [list(l) for l in chain.from_iterable(combinations(s, r) for r in range(len(s)+1))]


get_set_of_subsets = get_powerset



import re
from ._remove_non_alphanumeric import remove_non_alphanumeric as remove_non_alpha

_first_cap_re = re.compile('(.)([A-Z][a-z]+)')
_all_cap_re = re.compile('([a-z0-9])([A-Z])')


def convert_camel_to_snake(string, remove_non_alphanumeric=True, ignore_errors=False):
	"""
	converts CamelCase to snake_case
	:type string: str
	:rtype: str
	"""
	try:
		string = string.strip()

		if remove_non_alphanumeric:
			string = remove_non_alpha(string, replace_with='_', keep_underscore=True)

		s1 = _first_cap_re.sub(r'\1_\2', string)
		result = _all_cap_re.sub(r'\1_\2', s1).lower()
		result = re.sub(pattern='\s*_+', repl="_", string=result)
		return result
	except Exception as e:
		if ignore_errors:
			return string
		else:
			raise e



import datetime as dt
from datetime import date


FORMATS = (
	'%Y-%m-%d', '%Y/%m/%d', '%Y.%m.%d', '%Y%m%d',
	'%Y-%m', '%Y/%m', '%Y.%m', '%Y',
	'%y-%m-%d', '%y/%m/%d', '%y.%m.%d',
	'%b %d, %Y', '%b %d, %Y', '%B %d, %Y', '%B %d %Y', '%m/%d/%Y', '%m/%d/%y', '%b %Y', ' %B%Y', '%b %d,%Y'
)

def find_date_and_format(x, format=None):
	"""
	:type x: str or int or date
	:type format: NoneType or str
	:rtype: NoneType or date
	"""
	if format is None:
		if isinstance(x, str):
			formats = FORMATS
		elif isinstance(x, int):
			formats = ['%Y%m%d']
			x = str(x)
		else:
			raise TypeError(f'type {type(x)} is not supported!')
	else:
		formats = [format]

	for _format in formats:
		try:
			return dt.datetime.strptime(x, _format).date(), _format
		except ValueError:
			pass

	return None, None


def find_date_format(x, format=None):
	"""
	:type x: str or int or date
	:type format: NoneType or str
	:rtype: NoneType or date
	"""
	return find_date_and_format(x=x, format=format)[1]


def parse_date(x, format=None):
	"""
	:type x: str or int or date
	:type format: NoneType or str
	:rtype: NoneType or date
	"""
	return find_date_and_format(x=x, format=format)[0]



import pandas as pd
import math
from random import shuffle

class UpSampler:
	def __init__(self, fill_na=None, max_repeats=None):
		#self._sm = SMOTE(**kwargs)
		self._x_cols = None
		self._y_col = None
		self._fill_na = fill_na
		self._up_sampling_table = None
		self._max_repeats = max_repeats

	def fit(self, X, y):
		self._x_cols = X.columns
		self._y_col = y.name

		if self._fill_na is not None:
			y = y.fillna(self._fill_na)

		x_groups = {name: x for name, x in X.groupby(by=y)}
		#y_groups = {name: x for name, x in y.groupby(by=y)}
		group_names = list(x_groups.keys())
		up_sampling_table = pd.DataFrame({
			'name': group_names,
			'population': [x_groups[name].shape[0] for name in group_names]
		})
		num_groups = len(group_names)
		up_sampling_table['total_pop_according_to_groups'] = up_sampling_table['population'] * num_groups
		self._max_pop = up_sampling_table['total_pop_according_to_groups'].max()
		up_sampling_table['max_population'] = self._max_pop

		self._group_pop = int(self._max_pop / num_groups)
		up_sampling_table['ideal_group_pop'] = self._group_pop
		up_sampling_table['num_repeats_required'] = up_sampling_table['ideal_group_pop'] / up_sampling_table[
			'population']
		self._up_sampling_table = up_sampling_table
		self._repeats = up_sampling_table.set_index('name').to_dict()['num_repeats_required']
		if self._max_repeats is not None:
			self._repeats = {key:min(value, self._max_repeats) for key, value in self._repeats.items()}


	def fit_upsample(self, X, y):
		self.fit(X=X, y=y)
		return self.upsample(X=X, y=y)

	def upsample(self, X, y):
		"""
		:type X: pd.DataFrame
		:type y: pd.Series
		:rtype: tuple
		"""
		order = list(range(X.shape[0]))
		shuffle(order)
		X.index = order
		y.index = order

		if self._fill_na is not None:
			y = y.fillna(self._fill_na)



		x_groups = [
			pd.concat([x.sort_index()]*math.ceil(self._repeats[name])).head(self._group_pop)
			for name, x in X.groupby(by=y)
		]
		y_groups = [
			pd.concat([x.sort_index()] * math.ceil(self._repeats[name])).head(self._group_pop)
			for name, x in y.groupby(by=y)
		]

		return pd.concat(x_groups), pd.concat(y_groups)


from pandas import DataFrame
from ._DataContainer import get_display_function, DataContainer
from ._TrainingTestContainer import TrainingTestContainer
from ...time.progress import ProgressBar
from ._train_and_test import train_and_test


class ValidationContainer(DataContainer):
	def __init__(
			self, data, validation_indices, holdout_indices, folds, x_columns=None, y_column=None,
			sort_columns=None
	):
		super().__init__(data=data, x_columns=x_columns, y_column=y_column, sort_columns=sort_columns)
		self._validation_indices = validation_indices
		self._holdout_indices = holdout_indices
		self._folds = folds

	@property
	def validation_indices(self):
		return self._validation_indices

	@property
	def holdout_indices(self):
		return self._holdout_indices

	@property
	def validation(self):
		"""
		:rtype: DataFrame
		"""
		if self._validation_indices is None:
			return None
		else:
			return self._data.iloc[self._validation_indices]

	@property
	def holdout(self):
		"""
		:rtype: DataFrame
		"""
		if self._holdout_indices is None:
			return None
		else:
			return self._data.iloc[self._holdout_indices]

	@property
	def folds(self):
		"""
		:rtype: list[TrainingTestContainer]
		"""
		return self._folds

	@property
	def num_folds(self):
		return len(self._folds)

	def get_fold(self, n):
		"""
		:param n: a number between 1 and number of folds (it doesn't start at zero)
		:rtype: TrainingTestContainer
		"""
		return self.folds[n - 1]

	def evaluate_fold(
			self, fold_n, estimator, evaluation_type,
			estimator_id=None, x_columns=None, y_column=None,
			ignore_missing_y=False, ignore_missing_x=False,
			replace_missing_x='mean'
	):
		y_column = y_column or self._y_column
		if y_column is None:
			raise ValueError(f'y_column should be provided')
		x_columns = x_columns or self._x_columns
		if x_columns is None:
			x_columns = [col for col in self.columns if col != y_column]
		training_test_container = self.get_fold(n=fold_n)
		training_data = training_test_container.training_data
		test_data = training_test_container.test_data

		result = train_and_test(
			estimator=estimator, evaluation_type=evaluation_type,
			training_data=training_data, test_data=test_data,
			x_columns=x_columns, y_column=y_column,
			sort_columns=self._sort_columns,
			ignore_missing_y=ignore_missing_y,
			ignore_missing_x=ignore_missing_x,
			replace_missing_x=replace_missing_x
		)
		if estimator_id is not None:
			result = {'estimator_id': estimator_id, **result}

		return result

	def _generate_evaluation(self, estimators, evaluation_type, x_columns=None, y_column=None):
		"""
		a generator for the evaluate function
		:param estimators:
		:param evaluation_type:
		:param x_columns:
		:param y_column:
		:param aggregate:
		:rtype: Generator[dict] or list[dict]
		"""
		if not isinstance(estimators, (list, tuple, dict)):
			estimators = [estimators]

		if isinstance(estimators, (list, tuple)):
			estimators = {(i + 1): x for i, x in enumerate(estimators)}

		y_column = y_column or self._y_column
		if y_column is None:
			raise ValueError(f'y_column should be provided')
		x_columns = x_columns or self._x_columns
		if x_columns is None:
			x_columns = [col for col in self.columns if col != y_column]

		for i, fold in enumerate(self.folds):
			for estimator_id, estimator in estimators.items():
				yield {
					'kwargs': {
						'training_data': fold.training_data,
						'test_data': fold.test_data,
						'estimator': estimator,
						'evaluation_type': evaluation_type,
						'x_columns': x_columns,
						'y_column': y_column,
						'sort_columns': self._sort_columns
					},
					'estimator_id': estimator_id,
					'estimator_name': estimator.__class__.__name__,
					'fold': i + 1,
				}

	def _evaluate(self, estimators, evaluation_type, x_columns=None, y_column=None, aggregate=False, echo=1):
		records = []
		progress_bar = ProgressBar(total=len(estimators) * len(self.folds) + 1, echo=echo)
		progress_amount = 0

		for dictionary in self._generate_evaluation(
				estimators=estimators, evaluation_type=evaluation_type, x_columns=x_columns, y_column=y_column
		):
			fold_number = dictionary['fold']
			estimator_id = dictionary['estimator_id']
			progress_bar.show(amount=progress_amount, text=f'evaluating fold {fold_number} with estimator {estimator_id}')
			evaluation = train_and_test(**dictionary['kwargs'])
			records.append({
				'fold': fold_number,
				'estimator_id': estimator_id,
				'estimator_name': dictionary['estimator_name'],
				**evaluation
			})
			progress_amount += 1

		progress_bar.show(amount=progress_amount, text=f'aggregating results!')
		df = DataFrame.from_records(records)
		if aggregate:
			df = df.drop(columns='fold').groupby(
				['estimator_id', 'estimator_name']
			).agg(['mean', 'median', 'min', 'max', 'std'])
		progress_amount += 1
		progress_bar.show(amount=progress_amount, text=f'evaluation complete!')

		return df

	def evaluate_regression(self, estimators, x_columns=None, y_column=None, aggregate=False, echo=1):
		"""
		evaluates one or more regressors on all the folds and returns a dataframe of the results
		:param estimators: one or more regressors
		:param x_columns: the x columns (independent variables)
		:param y_column: the y column (dependent variable)
		:param aggregate: if True, the evaluation metrics will be aggregated over the folds using mean, median, std, min, max
		:param echo: to show progress or not
		:rtype: DataFrame
		"""
		return self._evaluate(
			evaluation_type='regression',
			estimators=estimators, x_columns=x_columns, y_column=y_column, echo=echo, aggregate=aggregate
		)

	def evaluate_classification(self, estimators, x_columns=None, y_column=None, aggregate=False, echo=1):
		"""
		evaluates one or more classifiers on all the folds and returns a dataframe of the results
		:param estimators: one or more classifiers
		:param x_columns: the x columns (independent variables)
		:param y_column: the y column (dependent variable)
		:param aggregate: if True, the evaluation metrics will be aggregated over the folds using mean, median, std, min, max
		:param echo: to show progress or not
		:rtype: DataFrame
		"""
		return self._evaluate(
			evaluation_type='classification',
			estimators=estimators, x_columns=x_columns, y_column=y_column, echo=echo, aggregate=aggregate
		)

	def display(self, p=None):
		display = get_display_function()

		if len(self.folds) > 0:
			for i, fold in enumerate(self.folds):
				fold.display(p=p, prefix=f'Fold {i + 1} ', function=display)
			if display is False:
				print('Holdout:')
				print(self.holdout)
			else:
				print('Holdout:')
				display(self.holdout)
		else:
			if display is False:
				print('Validation:')
				print(self.validation)
				print('Holdout:')
				print(self.holdout)
			else:
				print('Validation:')
				display(self.validation)
				print('Holdout:')
				display(self.holdout)



from ._general_exceptions import AtlanteanError, AmbiguityError


class DuplicationError(AtlanteanError):
	pass


class DataError(AtlanteanError):
	pass


class MissingColumnsError(DataError):
	pass


class MissingRowsError(DataError):
	pass


class DataAmbiguityError(DataError, AmbiguityError):
	pass


class LearningError(AtlanteanError):
	pass


class ClusteringError(LearningError):
	pass


class ModelNotFittedError(LearningError, RuntimeError):
	pass


class ModelAlreadyFittedError(LearningError, RuntimeError):
	pass


class EstimatorNotEvaluatedError(LearningError, RuntimeError):
	pass


class EvaluationError(LearningError, RuntimeError):
	pass



_COLOUR_NUMBERS = {
	'black': 0,
	'red': 1,
	'green': 2,
	'yellow': 3,
	'blue': 4,
	'purple': 5,
	'cyan': 6,
	'white': 7,
	'grey': 7
}


_TEXT_STYLES = {
	'normal': 0,
	'bold': 1,
	'underline': 2,
	'negative1': 3,
	'negative2': 5
}


def colour_text(text, text_colour=0, style=0, background_colour=None):
	if text_colour in _COLOUR_NUMBERS: text_colour = _COLOUR_NUMBERS[text_colour]
	if background_colour in _COLOUR_NUMBERS: background_colour = _COLOUR_NUMBERS[background_colour]
	if style in _TEXT_STYLES: style = _TEXT_STYLES[style]

	if background_colour is None and style == 0:
		return '\033[0;0m'+u"\u001b[38;5;" + str(text_colour % 256) + "m" + text + '\033[0;0m'
	else:
		return '\033[0;0m'+f'\033[{style};{text_colour%8+30};{background_colour%8+40}m{text}' + '\033[0;0m'




def convert_rgb_to_hsl(red, green, blue, max_value=1.0, min_value=0.0):
	red = (red - min_value) / max_value
	green = (green - min_value) / max_value
	blue = (blue - min_value) / max_value

	red = min(1.0, max(0.0, red/max_value))
	green = min(1.0, max(0.0, green))
	blue = min(1.0, max(0.0, blue))

	maximum = max(red, green, blue)
	minimum = min(red, green, blue)

	chroma = maximum - minimum

	if chroma == 0:
		hue = 0
	else:
		if red == maximum:
			segment = (green - blue) / chroma
			shift = 360/60 if segment < 0 else 0/60

		elif green == maximum:
			segment = (blue - red) / chroma
			shift = 120/60

		else:
			segment = (red - green) / chroma
			shift = 240/60

		hue = segment + shift




def _find_max_row(data, col):
	return data[data.index == data[col].idxmax()]

def _find_min_row(data, col):
	return data[data.index == data[col].idxmin()]

def select_extreme(data, col, group_by=None, extreme_type = 'max'):
	if extreme_type=='max':
		if group_by is None:
			return _find_max_row(data=data, col=col)
		else:
			return data.groupby(by = group_by).apply(lambda x: _find_max_row(data=x, col=col))
	else:
		if group_by is None:
			return _find_min_row(data=data, col=col)
		return data.groupby(by = group_by).apply(lambda x: _find_min_row(data=x, col=col))

def select_max(data, max_col, group_by=None):
	return select_extreme(data = data, col = max_col, group_by = group_by, extreme_type = 'max')

def select_min(data, min_col, group_by=None):
	return select_extreme(data = data, col = min_col, group_by = group_by, extreme_type = 'min')


from .SimpleController import SimpleController
from .JobController import JobController
from .ProcessController import ProcessController
from ._DEFAULT_VALUES import *


def Controller(
	model,
	time_unit=TIME_UNIT,
	max_cpu_count=MAX_CPU_COUNT,
	sleep_time=SLEEP_TIME,
	empty_count_limit=EMPTY_COUNT_LIMIT,
	max_sleep_time=MAX_SLEEP_TIME,
	n_jobs=N_JOBS
):
	if not isinstance(model, str):
		raise ValueError('controller_type should be a string')
	model = model.lower()

	if model in ['s', 'simple controller', 'simple']:
		return SimpleController(time_unit=time_unit)
	elif model in ['j', 'job controller', 'job', 'jobs']:
		return JobController(time_unit=time_unit, n_jobs=n_jobs)
	elif model in ['p', 'm', 'process controller', 'multiprocess controller', 'process', 'multiprocess']:
		return ProcessController(
			time_unit=time_unit, max_cpu_count=max_cpu_count,
			sleep_time=sleep_time, empty_count_limit=empty_count_limit, max_sleep_time=max_sleep_time
		)
	else:
		raise ValueError(f'unknown controller type: "{model}"')



from ._product import get_product
from ._duplicates import get_duplicates, remove_duplicates, has_duplicates
from ._flatten import flatten
from ._make_hashable import make_hashable
from .Counter import Counter


class List(list):
	def __init__(self, *args):
		if len(args) == 0:
			super().__init__()

		elif len(args) == 1:
			if isinstance(args[0], (list, tuple)):
				the_list = [
					self.__class__(element) if isinstance(element, list) and not isinstance(element, self.__class__)
					else element
					for element in args[0]
				]
				super().__init__(the_list)
			else:
				try:
					super().__init__(args[0])
				except TypeError:
					super().__init__([args[0]])

		else:
			the_list = [
				self.__class__(element) if isinstance(element, list) and not isinstance(element, self.__class__)
				else element
				for element in args
			]
			super().__init__(the_list)

	def __repr__(self):
		return f'List: [{", ".join(str(x) for x in self)}]'

	def __str__(self):
		return super().__str__()

	def __mul__(self, other):
		if isinstance(other, list):
			return self.__class__(get_product(self, other))
		else:
			return self * other

	def has_duplicates(self):
		"""
		returns True if list has duplicates
		:rtype: bool
		"""
		return has_duplicates(self)

	def remove_duplicates(self):
		"""
		removes the duplicates from list
		:rtype: List
		"""
		return self.__class__(remove_duplicates(self))

	def get_duplicates(self):
		"""
		returns duplicate values in list
		:rtype: List
		"""
		return self.__class__(get_duplicates(self))

	def flatten(self):
		"""
		flattens a list
		:rtype: List
		"""
		return self.__class__(flatten(self))

	def count(self, obj=None):
		"""
		if obj is None, counts the number of elements, otherwise counts the number of obj
		:rtype: List or int
		"""
		if obj is None:
			counter = Counter(self)
			return self.__class__(counter.items())
		else:
			return super().count(obj)

	def __hashkey__(self):
		return tuple([make_hashable(x) for x in self])



from ._standardize_columns import standardize_columns as sc


def long_to_wide(data, index_by, widen_cols, value_cols=None, standardize_cols=True):
	"""
	:type data: DataFrame
	:type index_by: list[str]
	:type widen_cols: list[str]
	:type value_cols: list[str]
	:type standardize_cols: bool
	:rtype: DataFrame
	"""
	result = data.pivot_table(
		aggfunc=lambda x: x,
		index=index_by,
		columns=widen_cols,
		values=value_cols
	).reset_index()

	if standardize_cols:
		sc(data=result, inplace=True)
	return result


tall_to_wide = long_to_wide



from pandas import concat, Series

from ...text import split_and_strip

# splits one column of a dataframe into multiple rows
# for example: the following dataframe
## number	label
## 1,2		something

# becomes:
## number


def split_rows(data, by_column, sep=',', split_column_name=None, keep_original=False):
	x = data[[by_column]]
	if split_column_name is None:
		split_column_name = by_column + '_split'
	split_data = concat([Series(row[by_column], split_and_strip(row[by_column], sep=sep)) for _, row in x.iterrows()]).reset_index()
	split_data.columns = [split_column_name, by_column]

	# if there is a column called by_column + _split in the original data there would be trouble
	result = split_data.merge(right=data, on=by_column, how='outer')

	if not keep_original:
		result.drop(axis=1, labels=by_column, inplace=True)
		result.rename(columns={split_column_name: by_column}, inplace=True)
	return result



from sklearn.linear_model import LinearRegression
from sklearn.tree import DecisionTreeClassifier
from pandas import concat
from ..validation import CrossValidation
from ._fill_na_and_flag import NaFillerFlagger
ROW_NUM_COLUMN = '_imputer_row_number_'


class ColumnImputer:
	def __init__(
			self, column, estimators, helper_columns=None, cross_validation=None,
			drop_na=False,
			main_metric='rmse'
	):
		"""
		:type estimator: LinearRegression or DecisionTreeClassifier or list
		:type column: str
		:type helper_columns: list[str] or NoneType
		:type cross_validation: CrossValidation or NoneType
		"""
		self._imputed_column = column
		self._helper_columns = helper_columns

		if not isinstance(estimators, (list, tuple)):
			estimators = [estimators]
		self._estimators = estimators

		self._cross_validation = cross_validation
		self._cross_validation_evaluation = None
		self._drop_na = drop_na
		self._main_metric = main_metric.lower()
		self._na_filler = None

		self._x_columns = None

	def fit(self, X):
		if self._imputed_column not in X.columns:
			raise KeyError(f'column "{self._imputed_column}" does not exist in the data!')
		if self._helper_columns is not None:
			missing_columns = [f'"{col}"' for col in self._helper_columns if col not in X.columns]
			if len(missing_columns) > 0:
				raise KeyError(f'missing columns: {", ".join(missing_columns)}')
		else:
			self._helper_columns = [col for col in X.columns if col != self._imputed_column]

		data = X[self._helper_columns + [self._imputed_column]]

		if self._drop_na:
			training_data = data.dropna(axis=0)
			training_y = training_data[self._imputed_column]
			training_x = training_data[self._helper_columns]

		else:
			training_data = data[data[self._imputed_column].notna()]

			training_y = training_data[self._imputed_column]
			training_x_with_missing = training_data[self._helper_columns]
			self._na_filler = NaFillerFlagger()
			training_x = self._na_filler.fit_transform(X=training_x_with_missing)

		if training_x.shape[1] == 0:
			raise RuntimeError('X has no columns!')
		elif training_x.shape[0] == 0:
			raise RuntimeError('X has no rows!')

		if self._estimators is None:
			raise RuntimeError('no estimator')

		if self._cross_validation is not None:
			training_data = training_x.copy()
			training_data[[self._imputed_column]] = training_y
			validation_container = self._cross_validation.split(data=training_data)
			self._cross_validation_evaluation = validation_container.evaluate_regression(
				estimators=self._estimators, x_columns=None, y_column=self._imputed_column
			)

		self._estimator.fit(training_x, training_y)
		self._x_columns = training_x.columns

	def _transform(self, X, return_y):
		data = X.copy()
		data[ROW_NUM_COLUMN] = range(data.shape[0])

		# divide data into two parts, the part that needs no help and the part that needs help
		not_missing = data[data[self._imputed_column].notna()]
		missing = data[data[self._imputed_column].isna()].copy()

		if not_missing.shape[0] == 0:
			raise ValueError(f'all values of the column "{self._imputed_column}" are missing!')

		# the helper columns should not have missing values, let's fix that
		missing_x = missing[self._helper_columns]
		if self._na_filler is not None:
			missing_x = self._na_filler.transform(missing_x)
			if missing_x.isnull().values.any():
				raise RuntimeError('transform did not remove missing values')

		if return_y:
			df_list = [not_missing[[ROW_NUM_COLUMN, self._imputed_column]]]
		else:
			df_list = [not_missing]

		if missing.shape[0] > 0:
			missing[self._imputed_column] = self._estimator.predict(missing_x)
			if return_y:
				df_list.append(missing[[ROW_NUM_COLUMN, self._imputed_column]])
			else:
				df_list.append(missing)

		all_data = concat(df_list).sort_values(ROW_NUM_COLUMN)

		if return_y:
			return all_data[self._imputed_column].values
		else:
			return all_data.drop(columns=ROW_NUM_COLUMN)

	def transform(self, X):
		return self._transform(X=X, return_y=False)



from .get_elapsed import get_elapsed, get_elapsed_months, get_elapsed_years
from .convert import convert
from .get_time import get_now, get_time
from .measure import measure
from .MeasurementSet import MeasurementSet
from .TimeMeasurement import TimeMeasurement
from .Timer import Timer
from ._Job import Job



from pandas import DataFrame
from ...collections import Series


def create_arguments(num_rounds, arguments):
	"""
	:type num_rounds: int
	:type arguments: dict[str, list or tuple or int or float]
	"""
	result = DataFrame({'_id_': [1]})
	for key, value in arguments.items():
		if isinstance(value, tuple) and len(value) == 2:
			values = Series(start=value[0], end=value[1], count=num_rounds, include_start=True, include_end=True).list

		elif isinstance(value, tuple) and len(value) == 3:
			values = Series(start=value[0], end=value[1], step=value[2], include_start=True, include_end=True).list

		elif isinstance(value, list):
			values = value

		elif not isinstance(value, (list, tuple)):
			values = [0.9 * value, value, 1.1 * value]

		data = DataFrame({key: values})
		data['_id_'] = 1
		result = result.merge(data, on='_id_')

	return result.drop(columns='_id_').reset_index(drop=True)



import numpy as np


def get_entropy(x):
    """
    Also known as Shanon Entropy
    Reference: https://en.wikipedia.org/wiki/Entropy_(information_theory)
    """
    unique, count = np.unique(x, return_counts=True, axis=0)
    probability = count/len(x)
    entropy = np.sum((-1)*probability*np.log2(probability))
    return entropy


def get_joint_entropy(x, y):
    """
    H(Y;X)
    Reference: https://en.wikipedia.org/wiki/Joint_entropy
    """
    xy = np.c_[x, y]
    return get_entropy(xy)


def get_conditional_entropy(y, x):
    """
    conditional entropy = Joint Entropy - Entropy of X
    H(Y|X) = H(Y;X) - H(X)
    Reference: https://en.wikipedia.org/wiki/Conditional_entropy
    """
    return get_joint_entropy(y, x) - get_entropy(x)


def get_information_gain(y, x):
    """
    Information Gain, I(Y;X) = H(Y) - H(Y|X)
    Reference: https://en.wikipedia.org/wiki/Information_gain_in_decision_trees#Formal_definition
    """
    return get_entropy(y) - get_conditional_entropy(y, x)



import pandas as pd
_min = min
_max = max


def min(data, axis=1, skip_na=True):
	"""

	:type data: pd.DataFrame
	:param axis:
	:return:
	"""
	if not skip_na:
		return data.min(axis=axis)

	def min_except_na(l):
		l2 = [x for x in l if x is not None]
		if len(l2)==0:
			return None
		else:
			return _min(l2)

	return data.apply(min_except_na, axis=axis)


def max(data, axis=1, skip_na=True):
	"""

	:type data: pd.DataFrame
	:param axis:
	:return:
	"""
	if not skip_na:
		return data.max(axis=axis)

	def max_except_na(l):
		l2 = [x for x in l if x is not None]
		if len(l2)==0:
			return None
		else:
			return _max(l2)

	return data.apply(max_except_na, axis=axis)


import multiprocess
from pandas import DataFrame
from time import sleep

from ...time import get_elapsed, get_now
from ...time.progress import ProgressBar

import atexit

from ._worker import worker
from ._TimeEstimate import MissingTimeEstimate
from .learning import LearningProject
from .learning import CrossValidationProject
from ._get_data_from_namespace import get_data_from_namespace, get_obj_from_namespace
from ._get_data_from_namespace import add_obj_to_namespace


class Processor:
	def __init__(self, time_unit='ms'):
		self._processes = {}
		self._manager = multiprocess.Manager()
		self._namespace = self._manager.Namespace()
		self._namespace_dir = set()
		self._estimators = {}

		self._to_do = self._manager.list()
		self._doing = self._manager.dict()
		self._done = self._manager.list()
		self._processed = []
		self._errors = []
		self._proceed_worker = self._manager.dict()
		self._worker_status = self._manager.dict()
		self._projects = {}

		self._tasks_by_id = {}
		self._time_unit = time_unit
		self._worker_id_counter = 0
		atexit.register(self.terminate)

		self._last_error_task = None

	def __repr__(self):
		lines = [
			'Processor',
			f'to-do: {len(self._to_do)}',
			f'doing: {len(self._doing)}',
			f'done: {len(self._done)}'
		]
		return '\n'.join(lines)

	@property
	def namespace(self):
		return self._namespace

	def add_data(self, data_id, data, overwrite=False):
		"""
		:type data_id: int or str
		:type data: DataFrame
		:type overwrite: bool
		"""
		self.add_obj(obj_type='data', obj_id=data_id, obj=data, overwrite=overwrite)
		self.add_obj(obj_type='columns', obj_id=data_id, obj=list(data.columns), overwrite=True)
		self.add_obj(obj_type='shape', obj_id=data_id, obj=data.shape, overwrite=True)

	def add_obj(self, obj_type, obj_id, obj, overwrite=False):
		add_obj_to_namespace(
			namespace=self.namespace, obj_type=obj_type, obj_id=obj_id, obj=obj, overwrite=overwrite
		)
		self._namespace_dir.add(f'{obj_type}_{obj_id}')

	@property
	def obj_directory(self):
		return self._namespace_dir

	def get_data(self, data_id=None, data_slice_id=None):
		"""
		:type data_id: int or str
		:type data_slice_id: int or str
		:rtype: DataFrame
		"""
		return get_data_from_namespace(namespace=self.namespace, data_id=data_id, data_slice_id=data_slice_id)

	def get_obj(self, obj_type, obj_id):
		"""
		:type obj_id: str
		:rtype: object
		"""
		return get_obj_from_namespace(namespace=self.namespace, obj_type=obj_type, obj_id=obj_id)

	def get_data_columns(self, data_id):
		"""
		:type data_id: int or str
		:rtype: list
		"""
		return getattr(self._namespace, f'columns_{data_id}')

	def get_data_shape(self, data_id):
		"""
		:type data_id: int or str
		:rtype: tuple
		"""
		return getattr(self._namespace, f'shape_{data_id}')

	def generate_worker_id(self):
		self._worker_id_counter += 1
		return f'worker_{self._worker_id_counter}'

	@property
	def projects(self):
		"""
		:rtype: dict[str, LearningProject or CrossValidationProject]
		"""
		return self._projects

	def add_project(self, project):
		"""
		:type project: LearningProject
		"""
		if project.name in self._projects:
			raise RuntimeError(f'project {project.name} already exists.')
		self._projects[project.name] = project
		project._processor = self

	def add_worker(self):
		"""
		:rtype: multiprocess.Process
		"""
		worker_id = self.generate_worker_id()
		process = multiprocess.Process(
			target=worker,
			kwargs={
				'worker_id': worker_id,
				'namespace': self._namespace,
				'to_do': self._to_do,
				'doing': self._doing,
				'done': self._done,
				'proceed': self._proceed_worker,
				'status': self._worker_status
			}
		)
		self._processes[worker_id] = process
		process.start()
		return process

	def add_workers(self, num_workers):
		"""
		:type num_workers: int
		"""
		self.process_done_tasks()
		for i in range(num_workers):
			self.add_worker()

	def create_cross_validation_project(
			self, name, y_column, problem_type, x_columns=None, time_unit='ms',
			evaluation_function=None, main_metric=None, lowest_is_best=None, best_score=None,
			scoreboard=None
	):
		"""

		:type 	name: str
		:type 	y_column: str

		:type 	problem_type: str
		:param 	problem_type: either regression or classification

		:type 	x_columns: str

		:type 	time_unit: str
		:param 	time_unit: s, ms, etc.


		:type 	evaluation_function: callable
		:param 	evaluation_function: 	a function that gets predicted and actual and
										produces a dictionary of values such as {'rmse': ...} for regression
										or {'f1_score': ...} for classification

		:param 	main_metric: 	the main metric used for comparison, it should exist as one of the keys
								in the result produced by evaluation_function

		:param 	lowest_is_best: usually True for regression (unless a weird metric is used) and False for classification
		:param 	best_score: usually 0 for regression and 1 for classification

		:type 	scoreboard: Scoreboard
		:param 	scoreboard: a Scoreboard object that keeps score of all estimators, can be added later too

		:rtype: CrossValidationProject
		"""
		project = CrossValidationProject(
			processor=self,
			name=name, y_column=y_column, problem_type=problem_type, x_columns=x_columns, time_unit=time_unit,
			evaluation_function=evaluation_function, main_metric=main_metric, lowest_is_best=lowest_is_best,
			best_score=best_score, scoreboard=scoreboard
		)
		return project

	def receive_to_do(self, project_name=None, num_tasks=None, echo=True, process_done_tasks=True, **kwargs):
		"""

		:param project_name: name of the project
		:type  project_name: str

		:param num_tasks: number of tasks to load from the project
		:type  num_tasks: int

		:param echo:
		:type  echo: bool

		:type  process_done_tasks: bool

		:return:
		"""
		if process_done_tasks:
			self.process_done_tasks()

		if project_name is None:
			for project_name in self.projects.keys():
				self.receive_to_do(
					project_name=project_name, num_tasks=num_tasks, echo=echo,
					process_done_tasks=False, **kwargs
				)

		else:
			project = self.projects[project_name]
			project.produce_tasks(ignore_error=True, echo=echo)
			if num_tasks is not None:
				if num_tasks > project.to_do_count:
					# fill to do list in project
					num_of_new_tasks = num_tasks - project.to_do_count

					# but cannot get more than what is available
					num_of_new_tasks = min(num_of_new_tasks, project.new_count)

					project.fill_to_do_list(num_tasks=num_of_new_tasks, **kwargs)

			loaded_count = 0
			while project.to_do_count > 0:
				task = project.pop_to_do()
				self._to_do.append(task)
				loaded_count += 1

			if echo:
				print(f'{loaded_count} loaded from project {project_name}')

	def process_done_tasks(self, ignore_errors=False, echo=True):
		processed_count = {}
		while True:
			try:
				task = self._done.pop(0)

				if task.status == 'done':
					self.projects[task.project_name].add_time_estimate(task=task)
					project = self.projects[task.project_name]
					project.add_done_task(task=task)
					if task.project_name not in processed_count:
						processed_count[task.project_name] = 1
					else:
						processed_count[task.project_name] += 1
				elif task.status == 'error':
					if not ignore_errors:
						self._last_error_task = task
						print(f'trace:{task._errors[0][1]}')
						raise task._errors[0][0]
				else:
					raise RuntimeError(f'do not know what to do with status: {task.status}')

				self._processed.append(task)

			except IndexError:
				break
		if echo:
			for project_name, number in processed_count.items():
				print(f'{number} tasks from project {project_name} processed.')

	def get_time_estimate(self, task):
		return self.projects[task.project_name].get_time_estimate(task=task)

	def count_to_do(self):
		return len(self._to_do) + len(self._doing)

	def count_done(self):
		return len(self._processed) + len(self._done)

	def get_to_do_time(self):
		if self.count_to_do() == 0:
			return 0

		total = 0
		for task in self._doing.values():
			estimate = self.get_time_estimate(task=task)
			if estimate == MissingTimeEstimate():
				return estimate
			total += estimate

		for task in self._to_do:
			estimate = self.get_time_estimate(task=task)
			if estimate == MissingTimeEstimate():
				return estimate
			total += self.get_time_estimate(task=task)

		return total

	def get_done_time(self):
		total = 0

		for task in self._processed:
			estimate = self.get_time_estimate(task=task)
			if estimate == MissingTimeEstimate():
				return estimate
			total += estimate

		for task in self._done:
			estimate = self.get_time_estimate(task=task)
			if estimate == MissingTimeEstimate():
				return estimate
			total += estimate

		return total

	def _get_worker_count_string(self):
		active = 0
		idle = 0
		terminated_or_ended = 0
		for worker_status in self._worker_status.values():
			if worker_status in {'started', 'active'}:
				active += 1
			elif worker_status == 'idle':
				idle += 1
			elif worker_status in {'ended', 'terminated'}:
				terminated_or_ended += 1
		result = []
		if active > 0:
			result.append(f'{active} active{", " if idle + terminated_or_ended > 0 else ""}')

		if idle > 0:
			result.append(f'{idle} idle{", " if terminated_or_ended > 0 else ""}')
		if terminated_or_ended > 0:
			result.append(f'{terminated_or_ended} ended')

		return ' '.join(result)

	def get_worker_count(self):
		return len(self._processes)

	def _update_progress_bay_by_count(self, progress_bar, next_line):
		to_do_count = self.count_to_do()
		done_count = self.count_done()
		progress_bar.set_total(to_do_count + done_count)
		progress_bar.show(
			amount=done_count,
			text=f'tasks: {done_count} / {to_do_count + done_count} | workers: {self._get_worker_count_string()}',
			next_line=next_line
		)
		return to_do_count

	def _update_progress_bar(self, progress_bar, next_line):
		self.process_done_tasks(echo=False)

		# if there is any time estimate use time to show progress

		# try with time:
		to_do_time = self.get_to_do_time()
		if to_do_time == MissingTimeEstimate():
			return self._update_progress_bay_by_count(progress_bar=progress_bar, next_line=next_line)

		done_time = self.get_done_time()
		if done_time == MissingTimeEstimate():
			return self._update_progress_bay_by_count(progress_bar=progress_bar)

		progress_bar.set_total(to_do_time + done_time)
		to_do_count = self.count_to_do()
		done_count = self.count_done()
		progress_bar.show(
			amount=done_time,
			text=f'tasks: {done_count} / {to_do_count + done_count} | workers: {self._get_worker_count_string()}',
			next_line=next_line

		)
		return to_do_count

	def show_progress(self, time_limit=None, time_unit='s'):
		if len(self._processes) == 0:
			raise RuntimeError('there are no workers')
		start_time = get_now()
		progress_bar = ProgressBar(total=100)
		progress_bar.show(amount=0)
		try:
			while True:
				to_do_count = self._update_progress_bar(progress_bar=progress_bar, next_line=False)
				if to_do_count == 0:
					progress_bar.set_total(total=100)
					progress_bar.show(amount=100, text=f'workers: {self._get_worker_count_string()}')
					break
				if time_limit is not None and get_elapsed(start=start_time, unit=time_unit) > time_limit:
					print()
					break
				sleep(0.1)

		except KeyboardInterrupt:
			self._update_progress_bar(progress_bar=progress_bar, next_line=True)

	@property
	def worker_status_table(self):
		return DataFrame.from_records([
			{'id': worker_id, 'status': worker_status}
			for worker_id, worker_status in self._worker_status.items()
		])

	@property
	def tasks(self):
		"""
		:rtype: list[TrainingTestTask]
		"""
		d = {}
		for task in self._processed:
			d[task.name] = task
		for task in self._done:
			d[task.name] = task
		for task in self._doing.values():
			d[task.name] = task
		for task in self._to_do:
			d[task.name] = task
		return list(d.values())

	@property
	def task_table(self):
		return DataFrame.from_records([
			task.record
			for task in self.tasks
		])

	def stop(self, worker_id=None):
		if worker_id is not None:
			self._proceed_worker[worker_id] = False

		else:
			for _worker_id in self._processes.keys():
				self.stop(worker_id=_worker_id)

		return self._done

	def terminate(self, worker_id=None, echo=1):
		self.stop(worker_id=worker_id)
		if worker_id is None:
			sleep(1)

		if worker_id is not None:
			if worker_id not in self._processes:
				raise KeyError(f'worker {worker_id}')

			self._processes[worker_id].terminate()
			if worker_id in self._doing:
				self._to_do.append(self._doing[worker_id])
				del self._doing[worker_id]
			del self._processes[worker_id]
			if self._worker_status[worker_id] != 'ended':
				self._worker_status[worker_id] = 'terminated'
				if echo:
					print(f'worker {worker_id} terminated!')
			else:
				if echo:
					print(f'worker {worker_id} already ended work.')

		else:
			worker_ids = list(self._processes.keys())
			for _worker_id in worker_ids:
				self.terminate(worker_id=_worker_id, echo=echo)

		return self._done



from ..parallel_computing import Processor
from ..parallel_computing import CrossValidationProject
from ..validation import CrossValidation, TimeSeriesValidation, Scoreboard
from ..validation import EstimatorRepository
import multiprocess
from pandas import DataFrame


class AutoML:
	def __init__(self):
		"""

		"""
		self._processor = Processor()
		self._estimator_repositories = {}

	@property
	def projects(self):
		"""
		:rtype: dict[str, CrossValidationProject]
		"""
		return self.processor.projects

	@staticmethod
	def get_cpu_count():
		return multiprocess.cpu_count()

	def get_worker_count(self):
		return self.processor.get_worker_count()

	@property
	def processor(self):
		"""
		:rtype: Processor
		"""
		return self._processor

	@property
	def estimator_repositories(self):
		"""
		:rtype: dict[str, EstimatorRepository]
		"""
		return self._estimator_repositories

	@property
	def score_boards(self):
		"""
		:rtype: dict[str, Scoreboard]
		"""
		return self._score_boards

	def add_validation(self, project_name, problem_type, y_column, data, validation, main_metric=None, random_state=None):
		"""
		:param 	project_name: the name or identification for a problem, should be unique
		:type	project_name: str or int

		:param 	problem_type: regression or classification
		:type 	problem_type: str

		:param 	y_column: name of the y column
		:type 	y_column: str

		:param	data: data used for the problem
		:type	data: DataFrame

		:param	validation: a cross validation or time-series validation skeleton
		:type 	validation: CrossValidation or TimeSeriesValidation

		:type 	random_state: int or NoneType
		"""

		self._projects[project_name] = CrossValidationProject(
			name=project_name, y_column=y_column, problem_type=problem_type, time_unit='ms', evaluation_function=None,
			main_metric=main_metric
		)

		validation_container = validation.split(data=data, random_state=random_state)

		estimator_repository = EstimatorRepository()
		self._estimator_repositories[project_name] = estimator_repository

		# add validation as a new problem
		validation_problem_id = f'{project_name}_validation'

		self._score_boards[validation_problem_id] = Scoreboard(
			estimator_repository=estimator_repository,
			main_metric=main_metric, lowest_is_best=lowest_is_best, best_score=best_score
		)

		self.processor.add_project(
			project_name=validation_problem_id, problem_type=problem_type, y_column=y_column
		)

		for i, fold in enumerate(validation_container.folds):
			data_id = f'fold_{i + 1}'
			self.processor.add_data(
				project_name=validation_problem_id,
				data_id=data_id,
				training_data=fold.training_data,
				test_data=fold.test_data
			)
			self.score_boards[project_name].add_training_test_id(training_test_id=data_id)

		# add final test as a new problem
		final_test_problem_id = f'{project_name}_final_test'

		self.processor.add_project(
			project_name=final_test_problem_id, problem_type=problem_type, y_column=y_column
		)

		self.processor.add_data(
			project_name=final_test_problem_id, data_id=f'final_test',
			training_data=validation_container.validation,
			test_data=validation_container.holdout
		)

	def add_models(self, problem_id, estimator, kwargs):
		self.estimator_repositories[problem_id].append(estimator=estimator, estimator_arguments=kwargs)

	def add_workers(self, num_workers=1):
		self.processor.add_workers(num_workers=num_workers)







import traceback
from ..time import get_elapsed, get_now


class TaskException:
	def __init__(self, task_id, exception, trace, worker_id):
		self._task_id = task_id
		self._exception = exception
		self._trace = trace
		self._worker_id = worker_id
		self._time = get_now()

	@property
	def exception(self):
		"""
		:rtype: Exception
		"""
		return self._exception

	@property
	def traceback(self):
		"""
		:rtype: str
		"""
		return self._trace

	@property
	def task_id(self):
		return self._task_id

	@property
	def time(self):
		return self._time

	@property
	def record(self):
		return {
			'task_id': self.task_id,
			'time': self.time,
			'exception': self.exception
		}


class Task:
	def __init__(self, function, task_id, args, kwargs, time_unit, cpu_count):
		self._function = function
		self._id = task_id
		self._args = args
		self._kwargs = kwargs
		self._time_unit = time_unit
		self._creation_time = get_now()
		self._cpu_count = cpu_count

	@property
	def id(self):
		return self._id

	@property
	def creation_time(self):
		return self._creation_time

	@property
	def time_unit(self):
		return self._time_unit

	@property
	def cpu_count(self):
		return self._cpu_count

	def do(self, worker_id=None):
		"""
		:rtype: Outcome
		"""
		start_time = get_now()
		try:
			if self._args is None:
				if self._kwargs is None:
					result = self._function()
				else:
					result = self._function(**self._kwargs)
			else:
				if self._kwargs is None:
					result = self._function(*self._args)
				else:
					result = self._function(*self._args, **self._kwargs)
		except Exception as exception:
			trace = traceback.format_exc()
			print(f'task {self.id} exception')
			traceback.print_exc()
			return TaskException(task_id=self.id, exception=exception, trace=trace, worker_id=worker_id)

		else:
			end_time = get_now()
			return Outcome(
				task=self, result=result, worker_id=worker_id,
				start_time=start_time, end_time=end_time
			)


class Outcome:
	def __init__(self, task, result, start_time, end_time, worker_id):
		"""
		:type task: Task
		:param result:
		:param start_time:
		:param end_time:
		:param worker_id:
		"""
		self._result = result
		self._creation_time = task.creation_time
		self._time_unit = task.time_unit
		self._task_id = task.id
		self._start_time = start_time
		self._end_time = end_time
		self._elapsed = get_elapsed(start=self._start_time, end=self._end_time, unit=self._time_unit)

		self._worker_id = worker_id
		self._signature = {'function_name': task._function.__name__, 'elapsed': self.elapsed}

		if task._args is not None:
			for i, arg in enumerate(task._args):
				if isinstance(arg, (int, float, str)):
					self._signature[f'arg_{i + 1}'] = arg
		if task._kwargs is not None:
			for key, value in task._kwargs.items():
				if isinstance(value, (int, float, str)):
					self._signature[key] = value

	@property
	def task_id(self):
		return self._task_id

	@property
	def elapsed(self):
		return self._elapsed

	@property
	def timestamp_record(self):
		return {
			'task_id': self._task_id,
			'worker_id': self._worker_id,
			'creation_time': self._creation_time,
			'start_time': self._start_time,
			'end_time': self._end_time,
			'elapsed': self.elapsed
		}

	@property
	def signature(self):
		return self._signature

	@property
	def result(self):
		return self._result



from pandas import DataFrame
from .ProgressBar import ProgressBar
_map = map


def map(function, iterable, percent=True, bar=True, time=True, text=''):
	"""
	:type function: function
	:type iterable: iterable
	:type progress_step: int
	:type percent: bool
	:type bar: bool
	:type time: bool
	:type text: str
	:rtype: iterable
	"""
	def _func(x, _progress, _progress_bar):
		result = function(x)
		_progress['amount'] += 1
		if _progress['max_step']>1:
			if _progress['amount'] % _progress['step']==0 or _progress['amount']>=total:
				_progress_bar.show(amount=_progress['amount'], percent=percent, bar=bar, time=time, text=text)


		return result

	total = len(iterable)
	progress = {'amount':0, 'step':1}
	progress_bar = ProgressBar(total=total)

	return _map(lambda x: _func(x=x, _progress=progress, _progress_bar=progress_bar), iterable)


def apply(function, data, progress_step=None, percent=True, bar=True, time=True, text=''):
	"""
	:type function: function
	:type data: DataFrame
	:type progress_step: int
	:type percent: bool
	:type bar: bool
	:type time: bool
	:type text: str
	:rtype: DataFrame
	"""
	def _func(x, _progress, _progress_bar):
		"""
		:type progress: dict[str, int]
		:type progress_bar: ProgressBar
		"""
		result = function(x)
		_progress['amount'] += 1
		if _progress['amount'] % _progress['step']==0 or _progress['amount']>=total:
			_progress_bar.show(amount=_progress['amount'], percent=percent, bar=bar, time=time, text=text)
		return result
	total = data.shape[0]
	if progress_step is None:
		progress_step = max(round(total/10),1) # we don't want progress step to be 0
	progress = {'amount': 0, 'step': 1, 'max_step': progress_step}
	progress_bar = ProgressBar(total=total)

	return data.apply(func=lambda x: _func(x=x, _progress=progress, _progress_bar=progress_bar))



from numpy import where, nan, int64
from pandas import to_numeric


def as_int(x):
	return where(x.astype(str).str.isnumeric(), x.astype(str).astype(int64, errors='ignore'), nan)


def as_numeric(x):
	return to_numeric(x, errors='coerce')


from pyspark.sql import DataFrame as SparkDF
from pandas import DataFrame as PandasDF
from pandas import concat
from pyspark.sql import functions as f
from multiprocessing.pool import ThreadPool
from itertools import chain
from ...collections import OrderedSet



def union(*args, n_jobs):
	"""
	takes multiple spark DataFrames and produces the union
	:type args: SparkDF or PandasDF or list[SparkDF] or list[PandasDF]
	:type n_jobs: int
	:rtype: DataFrame
	"""
	if len(args) == 1 and isinstance(args[0], list):
		args = args[0]

	if all([isinstance(x, SparkDF) for x in args]):
		global_data_type = 'spark_df'
	elif all([isinstance(x, PandasDF) for x in args]):
		global_data_type = 'pandas_df'
	else:
		raise TypeError('All dataframes should be the same type of DataFrame.')

	def _get_columns(x):
		if x is None:
			return []
		else:
			return list(x.columns)

	def _add_columns(_data, _all_columns):
		"""
		adds empty columns to _data for the columns it is missing from among all_columns
		:type _data: DataFrame
		:rtype: DataFrame
		"""
		if _data is None:
			return None

		_missing_columns = [_col for _col in _all_columns if _col not in _data.columns]
		if global_data_type == 'spark_df':
			return _data.select(*_data.columns, *[f.lit(None).alias(col) for col in _missing_columns])
		elif global_data_type == 'pandas_df':
			_result = _data.copy()
			for col in _missing_columns:
				_result[col] = None
			return _result
		else:
			raise TypeError(f'Unsupported data type: {type(_data)}')

	if n_jobs > 1:
		pool = ThreadPool(n_jobs)
		column_lists = pool.map(_get_columns, args)
		all_columns = OrderedSet(chain(*column_lists))
		df_list = pool.map(lambda x: _add_columns(x, _all_columns=all_columns), args)
	else:
		pool = None
		column_lists = [_get_columns(x) for x in args]
		all_columns = OrderedSet(chain(*column_lists))
		df_list = [_add_columns(x, _all_columns=all_columns) for x in args]

	def _union_by_name(_df1, _df2):
		"""
		orders _df2 columns in the same way as _df1 to make union easy
		:type _df1: DataFrame
		:type _df2: DataFrame
		:rtype: DataFrame
		"""
		if _df1 is None or _df1 is False:
			return _df2
		elif _df2 is None or _df2 is False:
			return _df1

		if global_data_type == 'spark_df':
			# make sure columns are the same type, otherwise cast them into string
			dtypes1 = {col: dtype for col, dtype in _df1.dtypes}
			dtypes2 = {col: dtype for col, dtype in _df2.dtypes}

			common_columns = set(dtypes1.keys()).intersection(dtypes2.keys())
			different_columns = []
			for col in common_columns:
				t1 = dtypes1[col].lower()
				t2 = dtypes2[col].lower()
				if t1 != t2:
					if t1 == 'string' or t2 == 'string' or t1.startswith('array') or t2.startswith('array') or t1.startswith('struct') or t2.startswith('struct'):
						different_columns.append(col)
			other_columns1 = [col for col in dtypes1.keys() if col not in different_columns]
			other_columns2 = [col for col in dtypes2.keys() if col not in different_columns]

			_df1 = _df1.select(*other_columns1, *[f.col(col).cast('string').alias(col) for col in different_columns])
			_df2 = _df2.select(*other_columns2, *[f.col(col).cast('string').alias(col) for col in different_columns])

			return _df1.union(_df2.select(*_df1.columns))

		elif global_data_type == 'pandas_df':
			return concat([_df1, _df2])

		else:
			raise TypeError(f'Unsupported types: {type(_df1)} and {type(_df2)}')

	def _reduce_in_half(_df_list):
		# unions half of the list with the other half and reduces the list to half its original size
		if len(_df_list) % 2 == 1:
			_new_list = _df_list + [None]
		else:
			_new_list = _df_list

		pairs = []
		for i in range(len(_new_list) // 2):
			pairs.append((_new_list[i*2], _new_list[i*2 + 1]))

		if pool is None:
			result = [_union_by_name(pair[0], pair[1]) for pair in pairs]
		else:
			result = pool.map(lambda pair: _union_by_name(pair[0], pair[1]), pairs)
		return result

	def _union(_df_list):
		"""
		:type _df_list: list[PandasDF] or list[SparkDF]
		:rtype: PandasDF or SparkDF
		"""
		result = _df_list.copy()
		while len(result) > 1:
			result = _reduce_in_half(result)
		return result[0]

	return _union(df_list)



from sklearn.decomposition import PCA
from sklearn.linear_model import LinearRegression


class PrincipalComponentModel:
	def __init__(self, model, pca):
		"""
		:type model: LinearRegression
		:type pca: PCA
		"""
		self._model = model
		self._pca = pca

	def fit(self, X, y, **kwargs):
		transformed_x = self._pca.fit_transform(X=X, y=y)
		self._model.fit(X=transformed_x, y=y, **kwargs)

	def predict(self, X):
		transformed_x = self._pca.transform(X=X)
		return self._model.predict(X=transformed_x)


class PrincipalComponentModelFactory:
	def __init__(self, model_class):
		self._model_class = model_class

	def compose_pca(self, **kwargs):
		"""
		:param kwargs:
		:rtype: PrincipalComponentModel
		"""
		pca_kwargs = {key[4:]: value for key, value in kwargs.items() if key.startswith('pca_')}
		model_kwargs = {key: value for key, value in kwargs.items() if not key.startswith('pca_')}
		pca = PCA(**pca_kwargs)
		model = self._model_class(**model_kwargs)

		return PrincipalComponentModel(model=model, pca=pca)



from pandas import DataFrame as PandasDF
from pyspark.sql import DataFrame as SparkDF


def assert_has_columns(df, columns):
    """
    :type df: PandasDF or SparkDF
    :type columns: str or list[str]
    """

    if isinstance(columns, str):
        columns = [columns]

    df_columns = list(df.columns)
    missing = [col for col in columns if col not in df_columns]
    if len(missing) > 0:
        message = f'Missing columns: "{", ".join(missing)}"'
        print(message)
        raise KeyError(message)
    else:
        return True



from ...exceptions import MissingArgumentError, AmbiguousArgumentsError
from ._get_cross_validation import get_cross_validation
from ._TrainingTestContainer import TrainingTestContainer

def get_training_test(
		data, sort_columns=None, id_columns=None, test_ratio=None, test_count=None, random_state=None
):
	"""
	:type data: DataFrame
	:type sort_columns: list[str]
	:type id_columns: list[str]
	:type test_ratio: float
	:type test_count: int
	:type random_state: int
	:rtype: TrainingTestContainer
	"""
	if test_ratio is None and test_count is None:
		raise MissingArgumentError('One of test_ratio or test_count should be provided!')
	elif test_ratio is not None and test_count is not None:
		raise AmbiguousArgumentsError('Only one of test_ratio or test_count should be provided!')
	elif test_ratio is None:
		test_ratio = test_count / data.shape[0]

	validation = get_cross_validation(
		data=data, num_splits=0, sort_columns=sort_columns, id_columns=id_columns,
		holdout_ratio=test_ratio, random_state=random_state,
	)

	return TrainingTestContainer(
		data=data, training_indices=validation.validation_indices, test_indices=validation.holdout_indices
	)



def get_feature_importances(model, columns, raise_error=True, **kwargs):
	try:
		importances = list(model.feature_importances_)
	except AttributeError:
		importances = None

	if importances is None:
		return None

	else:
		if len(columns) != len(importances):
			if raise_error:
				raise RuntimeError(f'number of columns: {len(columns)}, number of features: {len(importances)}')
			else:
				return None
		else:
			importance_dictionary = {column: importance for importance, column in zip(importances, columns)}

			for key, value in kwargs.items():
				importance_dictionary[key] = value

			sorted_dictionary = dict(sorted(importance_dictionary.items(), key=lambda item: -abs(item[1])))

			return sorted_dictionary




from .Timer import Timer
import functools


def measure(function, measurement_set, name, unit='ms'):
	"""
	:param callable function: function to be timed
	:type name: str
	:type unit: str
	:param .MeasurementSet_class.MeasurementSet measurement_set:
	:rtype: callable
	"""

	@functools.wraps(function)
	def wrapper(*args, **kwargs):
		timer = Timer(start_now=True, unit=unit)
		result = function(*args, **kwargs)
		timer.stop()
		measurement_set.add_measurement(name=name, timer=timer)
		return result
	wrapper.measurement_set = measurement_set

	return wrapper



from pandas import DataFrame
from pandas.api.types import is_numeric_dtype


def drop_single_value_columns(data, ignore=None, echo=True):
	"""
	:param 	data: a dataframe to be cleaned
	:type 	data: DataFrame
	:type 	echo: bool
	:rtype: DataFrame
	"""
	ignore = ignore or []
	data = data.copy()
	for column in data.columns:
		if column in ignore:
			continue
		uniques = data[column].unique()
		if len(uniques) == 1:
			if echo:
				print(f'dropping single value column: "{column}" with value "{uniques[0]}"')
			data.drop(column, inplace=True, axis=1)
	return data


def drop_columns_with_too_many_nas(data, na_ratio, ignore=None, echo=True):
	"""
	:param 	data: a dataframe to be cleaned
	:type 	data: DataFrame

	:param 	na_ratio: if the ratio of na's in a column (to the number of rows) is above this ratio, drop it
	:type 	na_ratio: float

	:type 	echo: bool

	:rtype: DataFrame
	"""
	data = data.copy()
	ignore = ignore or []
	num_na_in_columns = data.isnull().apply(sum, axis=0)
	if na_ratio >= 1:
		raise ValueError(f'na_ratio should be smaller than 1')

	for column in data.columns:
		if column in ignore:
			continue
		ratio = num_na_in_columns[column] / data.shape[0]
		if ratio >= na_ratio:
			if echo:
				print(f'dropping column "{column}" with na ratio of {round(ratio, 3)}')
				data.drop(columns=column, inplace=True)

	return data


def get_useless_column(data, column_a, column_b):
	"""
	if between column_a and column_b, all the information exists in column_a and we don't need column_b or vice versa
	:param 	data: the dataframe of question
	:type 	data: DataFrame
	:type 	column_a: str
	:type 	column_b: str
	"""

	column_a_uniques = data[[column_a]].drop_duplicates().shape[0]
	column_b_uniques = data[[column_b]].drop_duplicates().shape[0]

	combined_column_uniques = data[[column_a, column_b]].drop_duplicates().shape[0]

	if combined_column_uniques == column_a_uniques == column_b_uniques:
		# either column_a or column_b is useless
		# we prefer to drop non-numeric columns
		if is_numeric_dtype(data[column_a]):
			return column_b
		elif is_numeric_dtype(data[column_b]):
			return column_a
		else:
			# when all equal, second one gets the cut
			return column_b

	if combined_column_uniques == column_b_uniques:
		# column_a doesn't add any information
		return column_a

	elif combined_column_uniques == column_a_uniques:
		# column_b doesn't add any information
		return column_b

	else:
		return None


def drop_columns_with_no_additional_information(data, ignore=None, echo=True):
	useless_columns = set()
	columns = list(data.columns)

	"""
	example: 
	data = pd.DataFrame({
		'a': [4, 4, 4, 4, 5], 
		'b': [1, 2, 1, 1, 3], 
		'c': ['a', 'b', 'a', 'a', 'd'],
		'd': ['a', 'b', 'a', 'a', 'a']
	})
	all columns but b should be dropped
	"""
	ignore = ignore or []
	for i in range(len(columns) - 1):
		column_a = columns[i]
		if column_a in useless_columns or is_numeric_dtype(data[column_a]):
			continue
		else:
			for j in range(i + 1, len(columns)):
				column_b = columns[j]
				if column_b in useless_columns or is_numeric_dtype(data[column_b]):
					continue
				else:
					useless_column = get_useless_column(data=data, column_a=column_a, column_b=column_b)
					if useless_column is not None and useless_column not in ignore:
						useless_columns.add(useless_column)
						if echo:
							print(f'between "{column_a}" and "{column_b}", "{useless_column}" is useless.')

	return data.drop(columns=useless_columns)


def drop_bad_columns(data, na_ratio, ignore=None, echo=True):
	"""
	:param 	data: a dataframe to be cleaned
	:type 	data: DataFrame

	:param 	na_ratio: if the ratio of na's in a column (to the number of rows) is above this ratio, drop it
	:type 	na_ratio: float

	:param 	ignore: list of columns to be totally ignored (get a free pass)
	:type 	ignore: list[str] or set[str]
	:type 	echo: bool

	:rtype: DataFrame
	"""
	data = drop_single_value_columns(data, ignore=ignore, echo=echo)
	data = drop_columns_with_no_additional_information(data=data, ignore=ignore, echo=echo)
	return drop_columns_with_too_many_nas(data=data, na_ratio=na_ratio, ignore=ignore, echo=echo)



from ._make_hashable import make_hashable
import heapq as _heapq


class Counter:
	'''
	similar to the Counter class from collections but with the ability to count objects that are not hashable
	'''
	def __init__(self, l):
		self._counts = {}
		self._values = {}
		for x in l:
			hashable = make_hashable(x)
			if hashable in self._counts:
				self._counts[hashable] += 1
				self._values[hashable] = x
			else:
				self._counts[hashable] = 1
				self._values[hashable] = x

	def items(self):
		return [(self._values[key], count) for key, count in self._counts.items()]

	def most_common(self, n=None):
		'''
		List the n most common elements and their counts from the most
		common to the least.  If n is None, then list all element counts.

		>>> Counter('abcdeabcdabcaba').most_common(3)
		[('a', 5), ('b', 4), ('c', 3)]

		'''
		if n is None:
			sorted_counts = sorted(self._counts.items(), key=lambda x: x[1], reverse=True)
			return [(self._values[hashable], count) for hashable, count in sorted_counts]

		else:
			sorted_counts = _heapq.nlargest(n, self._counts.items(), key=lambda x: x[1])
			return [(self._values[hashable], count) for hashable, count in sorted_counts]

	def __repr__(self):
		string = ', '.join(f'{x}: {y}' for x, y in self.most_common())
		return 'Counter: {' + string + '}'



from ...time import get_elapsed
from ._get_data_from_namespace import get_data_from_namespace
from datetime import datetime
import traceback

class Task:
	def __init__(self, project_name, task_id):
		self._project_name = project_name
		self._status = 'new'
		self._worker_id = None
		self._errors = []
		self._starting_time = None
		self._ending_time = None
		self._id = task_id

	def __str__(self):
		return f'{self.id} ({self._status})'

	def __repr__(self):
		return f'{self.__class__.__name__}: {self})'

	@property
	def time_estimate_id(self):
		return 'generic_task'

	@property
	def id(self):
		if self._id is None:
			error = ValueError(f'task id is None!')
			self.add_error(error=error, trace=traceback.format_exc())
			raise error
		return self._id

	@property
	def project_name(self):
		return self._project_name

	@property
	def starting_time(self):
		return self._starting_time

	@property
	def ending_time(self):
		return self._ending_time

	def start(self):
		self._status = 'started'
		self._starting_time = datetime.now()

	def get_data_from_namespace(self, namespace, data_id=None, data_slice_id=None):
		return get_data_from_namespace(namespace=namespace, data_id=data_id, data_slice_id=data_slice_id)

	def do(self, namespace, worker_id):
		try:
			self.start()

			# placeholder for whatever needs to be done
			error = NotImplementedError(f'do() is not implemented for this task: {self.id}')
			self.add_error(error=error, trace=traceback.format_exc())
			raise error

			self.end(worker_id=worker_id)
		except Exception as error:
			self.add_error(error=error)

	def end(self, worker_id):
		ending_time = datetime.now()
		if self.status == 'error':
			error = RuntimeError('cannot change error status to done!')
			self.add_error(error=error, trace=traceback.format_exc())
			raise error
		elif self.status == 'started':
			self._status = 'done'
			self._worker_id = worker_id
		else:
			error = RuntimeError(f'unknown status: {self.status}')
			self.add_error(error=error, trace=traceback.format_exc())
			raise error

		self._ending_time = ending_time
		self._worker_id = worker_id

	def get_elapsed(self, unit='ms'):
		if self._starting_time is not None and self._ending_time is not None:
			return get_elapsed(start=self.starting_time, end=self.ending_time, unit=unit)
		else:
			return None

	@property
	def status(self):
		return self._status

	def add_error(self, error, trace=None):
		self._errors.append((error, trace))
		self._status = 'error'

	@property
	def errors(self):
		return self._errors

	def has_error(self):
		return len(self._errors) > 0

	def is_done(self):
		return self._status == 'done'






from pandas import DataFrame as PandasDF
from pyspark.sql import DataFrame as SparkDF
from pyspark.sql import functions as f
from ...functions import get_function_arguments


def apply_function_in_a_simple_way(data, function, new_column, return_type=None):
	"""
	:type data: PandasDF or SparkDF
	:type function: callable
	:type new_column: str
	:type return_type:
	"""
	args = get_function_arguments(function=function)
	missing_columns = [arg for arg in args if arg not in data.columns]
	if len(missing_columns) > 0:
		raise KeyError(f'Columns missing: "{", ".join(missing_columns)}"')

	if isinstance(data, SparkDF):
		if return_type is None:
			raise ValueError(f'return_type should be provided for pyspark DataFrame.')
		data = data.withColumn(
			new_column,
			f.udf(function, return_type)(*[f.col(arg) for arg in args])
		)

	elif isinstance(data, PandasDF):
		data = data.copy()
		def _function(row):
			return function(*[row[arg] for arg in args])
		data[new_column] = data.apply(lambda row: _function(row), axis=1)

	else:
		raise TypeError(f'data of type "{type(data)}" is not supported!')

	return data


def apply_function(data, function, new_column, return_type=None, unique_ratio=0.01, echo=0):
	"""
	takes a function and applies it to all rows of the data
	the keyword arguments of the function should match the columns of the data
	for example, if the function is: f(feature_1, feature_2, feature_3) --> feature_1 + feature_2 + feature_3
	apply function finds the three columns called feature_1, feature_2, feature_3 and runs the function
	on each row of the data and saves the result as values in a new column called new_column

	unique_ratio is used to decide if performance should be optimized or not.
	if ratio of unique cases to all cases (number of rows of data) is smaller than unique_ratio
	then the function will only be applied to unique cases and then the result will be joined with the original data

	:type data: PandasDF or SparkDF
	:type function: callable
	:type new_column: str
	:type unique_ratio: float
	:type echo: int or bool
	"""

	if unique_ratio == 0:
		return apply_function_in_a_simple_way(
			data=data, function=function, return_type=return_type, new_column=new_column
		)

	else:
		args = get_function_arguments(function=function)
		missing_columns = [arg for arg in args if arg not in data.columns]
		if len(missing_columns) > 0:
			raise KeyError(f'Columns missing: "{", ".join(missing_columns)}"')

		if isinstance(data, PandasDF):
			uniques = data[args].drop_duplicates()
			n_uniques = uniques.shape[0]
			n_total = data.shape[0]
		elif isinstance(data, SparkDF):
			uniques = data.select(*args).distinct()
			n_uniques = uniques.count()
			n_total = data.count()
		else:
			raise TypeError(f'data of type "{type(data)}" is not supported!')

		if n_uniques / n_total > unique_ratio:
			return apply_function_in_a_simple_way(
				data=data, function=function, return_type=return_type, new_column=new_column
			)

		else:
			if echo:
				print(f'applying the function only on {n_uniques} unique cases instead of {n_total} cases.')

			uniques = apply_function_in_a_simple_way(
				data=uniques, function=function, return_type=return_type, new_column=new_column
			)

			columns = list(data.columns) + [new_column]

			if isinstance(data, PandasDF):
				return uniques.merge(data, on=args, how='left')[columns]
			else:
				return uniques.join(data, on=args, how='left').select(*columns)



def flatten(l):
	"""
	flattens a list
	:param list l:
	:rtype: list
	"""
	flat_list = []
	for item in l:
		if isinstance(item, (list, tuple)):
			flat_list += flatten(item)
		else:
			flat_list.append(item)
	return flat_list



from pyspark.sql import DataFrame
from pyspark.ml.feature import VectorAssembler, StandardScaler
from .disassemble import disassemble
from ..testing import assert_has_columns


_SCALER_STATE = [
    '_columns', '_assembler', '_with_mean', '_with_sd', '_scaler', '_fitted'
]

class Scaler:
    def __init__(self, columns=None, with_mean=True, with_sd=True):
        """
        :type columns: list[str]
        :type with_mean: bool
        :type with_sd: bool
        """
        self._columns = columns
        self._assembler = None
        self._with_mean = with_mean
        self._with_sd = with_sd
        self._scaler = None
        self._fitted = False

    def __getstate__(self):
        return {
            name: getattr(self, name) for name in _SCALER_STATE
        }

    def __setstate__(self, state):
        for name, value in state.items():
            setattr(self, name, value)

    def fit(self, X):
        """
        :type X: DataFrame
        :rtype: Scaler
        """
        if self._columns is None:
            self._columns = X.columns
        else:
            assert_has_columns(X, columns=self._columns)

        self._assembler = VectorAssembler(inputCols=self._columns, outputCol='features')
        assembled = self._assembler.transform(X).select('features')
        self._scaler = StandardScaler(
            inputCol='features', outputCol='scaled_features',
            withMean=self._with_mean, withStd=self._with_sd
        ).fit(assembled)
        self._fitted = True
        return self

    def transform(self, X, return_vector=False):
        """
        :type X: DataFrame
        :type return_vector: bool
        :param return_vector: if True, the vector column will be returned
                            if False, the vector column will be disassembled
        :rtype: DataFrame
        """
        if not self._fitted:
            self.fit(X)

        assert_has_columns(X, columns=self._columns)

        assembled = self._assembler.transform(X).select(
            'features', *[col for col in X.columns if col not in self._columns]
        )
        df_scaled = self._scaler.transform(assembled).drop('features')
        if return_vector:
            return df_scaled
        else:
            names = [f'{col}_scaled' for col in self._columns]
            return disassemble(df_scaled, column='scaled_features', names=names, drop=True)

    fit_transform = transform



from pyspark.sql import DataFrame
from pyspark.ml.feature import VectorAssembler, PCA as SparkPCA
from ..testing import assert_has_columns
from .Scaler import Scaler
from .disassemble import disassemble


_PCA_STATE = [
	'_columns', '_scale', '_assembler', '_k', '_scaler', '_scale_with_means', '_scale_with_sd', '_fitted',
	'_pca_columns'
]


class PCA:
	def __init__(self, k=None, columns=None, scale=False, scale_with_means=True, scale_with_sd=True):
		"""

		:type k: int or NoneType
		:type columns: NoneType or list[str]
		:type scale: bool
		:type scale_with_means: bool
		:type scale_with_sd: bool
		"""
		self._columns = columns
		self._scale = scale
		self._assembler = None
		self._k = k

		self._scaler = None
		self._scale_with_means = scale_with_means
		self._scale_with_sd = scale_with_sd

		self._pca = None
		self._fitted = False
		self._pca_columns = None

	def __getstate__(self):
		return {
			name: getattr(self, name)
			for name in _PCA_STATE
		}

	def __setstate__(self, state):
		for name, value in state.items():
			setattr(self, name, value)

	def fit(self, X):
		"""
		:type X: DataFrame
		:rtype: PCA
		"""
		if self._columns is None:
			self._columns = X.columns
		else:
			assert_has_columns(X, columns=self._columns)

		if self._scale:
			self._scaler = Scaler(columns=self._columns, with_mean=self._scale_with_means, with_sd=self._scale_with_sd)
			scaled = self._scaler.fit_transform(X=X, return_vector=True)
			self._pca = SparkPCA(k=self._k, inputCol='scaled_features', outputCol='pca_features').fit(scaled)

		else:
			self._assembler = VectorAssembler(inputCols=self._columns, outputCol='features')
			assembled = self._assembler.transform(X).select('features')
			self._pca = SparkPCA(k=self._k, inputCol='features', outputCol='pca_features').fit(assembled)

		self._fitted = True
		return self

	def transform(self, X, return_vector=False):
		"""
		:type X: DataFrame
		:type return_vector: bool
		:param return_vector: True: the vector column will be returned. False: the vector column will be disassembled.
		:rtype: DataFrame
		"""
		if not self._fitted:
			self.fit(X=X)

		assert_has_columns(X, self._columns)

		if self._scale:
			scaled = self._scaler.transform(X=X, return_vector=True)
		else:
			scaled = self._assembler.transform(X).select(
				'features', *[col for col in X.columns if col not in self._columns]
			)

		transformed = self._pca.transform(scaled).drop('features', 'scaled_features')
		if return_vector:
			return transformed

		else:
			self._pca_columns = [f'pca_{i}' for i in range(1, self._k + 1)]
			return disassemble(transformed, column='pca_features', names=self._pca_columns, drop=True)

	fit_transform = transform

	@property
	def variance_ratios(self):
		"""
		:rtype: list[float]
		"""
		return list(self._pca.explainedVariance.toArray())

	@property
	def pca_columns(self):
		"""
		:rtype: list[str]
		"""
		return self._pca_columns

	@property
	def variance_ratio_by_column(self):
		"""
		:rtype: dict[str, float]
		"""
		variance_ratios = self.variance_ratios
		pca_columns = self.pca_columns
		if len(variance_ratios) != len(pca_columns):
			raise RuntimeError(f'There are {len(variance_ratios)} variance ratios for {len(pca_columns)} columns!')

		return {column: variance_ratio for column, variance_ratio in zip(pca_columns, variance_ratios)}



from pandas.api.types import is_numeric_dtype


def get_numeric_columns(data):
	return [col for col in data.columns if is_numeric_dtype(data[col])]


def get_na_columns(data, only_numeric=True):
	if only_numeric:
		data = data[get_numeric_columns(data=data)]

	d = dict(data.isnull().any(axis=0))
	return [k for k, v in d.items() if v]


def fill_na_and_flag(data, fill_with='median'):
	data = data.copy()
	na_columns = get_na_columns(data=data, only_numeric=True)

	for column in na_columns:
		na_column = f'{column}_is_null'
		data[na_column] = data[column].isnull() * 1
		if fill_with == 'median':
			replacement = data[column].median()
		elif fill_with == 'mean':
			replacement = data[column].mean()
		else:
			raise ValueError(f'unknown option fill_with="{fill_with}"')
		data[column].fillna(replacement, inplace=True)

	return data


class NaFillerFlagger:
	def __init__(self, fill_with='median'):
		fill_with = fill_with.lower()
		if fill_with not in ['mean', 'median']:
			raise ValueError(f'unknown option fill_with="{fill_with}"')
		self._fill_with = fill_with
		self._na_columns = None

	def fit(self, X):
		self._na_columns = get_na_columns(data=X, only_numeric=True)

	def transform(self, X):
		data = X.copy()
		for col in self._na_columns:
			na_column = f'{col}_is_null'
			data[na_column] = data[col].isnull() * 1
			if self._fill_with == 'median':
				replacement = data[col].median()
			elif self._fill_with == 'mean':
				replacement = data[col].mean()
			else:
				raise ValueError(f'unknown option fill_with="{self._fill_with}"')
			data[col].fillna(replacement, inplace=True)

		for col in X.columns:
			if col not in self._na_columns and data[col].isnull().any():
				if self._fill_with == 'median':
					replacement = data[col].median()
				elif self._fill_with == 'mean':
					replacement = data[col].mean()
				else:
					raise ValueError(f'unknown option fill_with="{self._fill_with}"')
				data[col].fillna(replacement, inplace=True)
		return data

	def fit_transform(self, X):
		self.fit(X=X)
		return self.transform(X=X)



from pandas import DataFrame
from ..evaluation import evaluate_classification, evaluate_regression
from ..wrangling import bring_to_front
from ...exceptions import ComparisonError


class Result:
	def __init__(self, actual, predicted, probabilities, data_type, problem_type):
		self._actual = actual
		self._predicted = predicted
		self._probabilities = probabilities
		self._performance = None
		self._data_type = data_type
		self._problem_type = problem_type

	@property
	def performance(self):
		if self._performance is None:
			if self._problem_type.lower()[0] == 'r':
				self._performance = evaluate_regression(actual=self._actual, predicted=self._predicted)
			else:
				self._performance = evaluate_classification(actual=self._actual, predicted=self._predicted)
		return self._performance

	@property
	def data(self):
		if self._probabilities is None:
			result = DataFrame({
				'actual': self._actual,
				'predicted': self._predicted
			})
		else:
			result = DataFrame({
				'actual': self._actual,
				'predicted': self._predicted,
				'prediction_probability': self._probabilities
			})
		result['result_type'] = self._data_type
		return bring_to_front(result, columns=['result_type'])


class TestResult(Result):
	def __init__(self, actual, predicted, probabilities, problem_type):
		super().__init__(
			actual=actual, predicted=predicted, probabilities=probabilities, data_type='test',
			problem_type=problem_type
		)


class TrainingResult(Result):
	def __init__(self, actual, predicted, probabilities, problem_type):
		super().__init__(
			actual=actual, predicted=predicted, probabilities=probabilities, data_type='training',
			problem_type=problem_type
		)



from ...collections import cumsum


def split_data(data, count_column, ratios=None, counts=None):
	split_cumsum = data[count_column].cumsum()

	total_sum = data[count_column].sum()

	if counts is None and ratios is not None:
		cumulative_ratios = cumsum(ratios)
		if round(cumulative_ratios[-1], 5) != 1:
			print('ratios', ratios)
			print('total sum', total_sum)
			print('cumulative_ratios', cumulative_ratios)
			raise ValueError('total of ratios is not 1')
		cumulative_ratios[-1] = 1
		cumulative_counts = [x * total_sum for x in cumulative_ratios]

	else:
		cumulative_counts = [round(x) for x in cumsum(counts)]
		if cumulative_counts[-1] != total_sum:
			raise ValueError(f'cumulative count {cumulative_counts[-1]} is not equal total_sum {total_sum} ')

	result = []
	for min_count, max_count in zip([0] + cumulative_counts[:-1], cumulative_counts):
		if min_count == max_count:
			result.append(None)
		else:
			subdata = data[(min_count < split_cumsum) & (split_cumsum <= max_count)]
			result.append(subdata)

	return result



import re


def remove_non_alphanumeric(s, replace_with='_', keep_underscore=True):
	s = str(s)
	# replace & with and
	s = re.sub(r'&+', '&', s)
	s = s.replace(' & ', ' and ')
	s = s.replace('_&_', '_and_')
	s = s.replace('&', ' and ')

	if keep_underscore:
		return re.sub(r'[\W]+', replace_with, s, flags=re.UNICODE)
	else:
		return re.sub(r'[\W_]+', replace_with, s, flags=re.UNICODE)


def remove_non_alphabetic(s):
	return ''.join([i for i in s if i.isalpha()])



import re
from ..numbers import FlexibleNumber, NumberPart
from .get_quarter import get_quarter
from .get_first_day_of_month import get_first_day_of_month
from .get_first_day_of_month import get_last_day_of_month
import datetime


class YearQuarter(FlexibleNumber):

	def __init__(self, x=None, year=None, quarter=None, date=None, parts=None, sep='-'):
		if type(x) is int:
			year = x // 10
			quarter = x % 10
		elif type(x) is str:
			x = str(x)
			year = int(re.findall(pattern='^\d+', string=x)[0][:4])
			quarter = int(re.findall(pattern='\d+$', string=x)[0][-1:])

		try:
			year = date.year
			quarter = get_quarter(date=date)
		except:
			pass

		try:
			year_part = NumberPart(value=year, base=None, digits=4)
			quarter_part = NumberPart(value=quarter, base=4, start=1)
		except:
			year_part = parts[0]
			quarter_part = parts[1]

		if sep is None:
			sep = '-'

		super().__init__(parts=[year_part, quarter_part], labels=['year', 'quarter'], sep=sep)
		self.adjust()

	@property
	def year(self):
		return self.get_value('year')

	@property
	def quarter(self):
		return self.get_value('quarter')

	def __str__(self):
		return f'{self.year}Q{self.quarter}'

	def __repr__(self):
		return self.__str__()

	@year.setter
	def year(self, year):
		self.set(value=int(year), label='year')

	@quarter.setter
	def quarter(self, quarter):
		self.set(value=int(quarter), label='quarter')

	def to_int(self):
		return self.year*100 + self.quarter

	def to_quarters(self):
		return self.get_total(label='quarter')

	def to_years(self):
		return self.get_total(label='year')

	@property
	def first_month(self):
		return self.last_month - 2

	@property
	def last_month(self):
		return self.quarter * 3

	@property
	def first_day(self):
		"""
		:rtype: datetime.date
		"""
		return get_first_day_of_month(year=self.year, month=self.first_month)

	@property
	def last_day(self):
		"""
		:rtype: datetime.date
		"""
		return get_last_day_of_month(year=self.year, month=self.last_month)



from collections import deque
import queue
import multiprocess
from pandas import DataFrame
from sys import stdout

from ._Task import Task, Outcome
from ._DEFAULT_VALUES import TIME_UNIT


class BaseController:
	def __init__(self, time_unit=TIME_UNIT):
		self._time_unit = time_unit
		self._to_do_queue = deque()
		self._done_queue = deque()
		self._tasks_status = dict()
		self._processed = dict()
		self._task_counter = 0
		self._tasks_status = dict()
		self._incomplete_task_ids = dict()
		self._worker_id_counter = 0

	def __repr__(self):
		return f'<{self.__class__.__name__} with to-do:{self._get_to_do_plus()}, done:{self.done_count}>'

	def _get_to_do_plus(self):
		return self.to_do_count + self.being_done_count

	@staticmethod
	def write(string, flush=True):
		"""
		:type string: str
		:type flush: bool
		"""
		stdout.write('\r' + string)
		if flush:
			stdout.flush()

	@property
	def system_cpu_count(self):
		return multiprocess.cpu_count()

	@property
	def to_do_queue(self):
		"""
		:rtype: deque[Task] or multiprocessing.Queue[Task]
		"""
		return self._to_do_queue

	@property
	def being_done_count(self):
		return 0

	@property
	def done_queue(self):
		"""
		:rtype: deque[Outcome] or multiprocessing.Queue[Outcome]
		"""
		return self._done_queue

	def _get_from_done_queue(self):
		"""
		:rtype: Outcome
		"""
		return self.done_queue.popleft()

	@property
	def processed_count(self):
		return len(self.processed)

	@property
	def to_do_count(self):
		return len(self.to_do_queue)

	@property
	def done_count(self):
		return len(self.done_queue)

	@property
	def processed(self):
		"""
		:rtype: dict[str, Outcome] or dict[int, Outcome]
		"""
		return self._processed

	def _create_task(self, function, args=None, kwargs=None, task_id=None, cpu_count=1):
		if task_id is None:
			task_id = self._task_counter + 1
		elif isinstance(task_id, int):
			if task_id <= self._task_counter:
				raise ValueError(f'task_id: {task_id} and it cannot be a number smaller than or equal to task counter')
		elif isinstance(task_id, str):
			if task_id in self._tasks_status:
				raise KeyError(f'task_id: "{task_id}" already exists!')
		else:
			raise TypeError(f'task_id of type {type(task_id)} is not acceptable.')

		if cpu_count == -1:
			cpu_count = self.system_cpu_count
		task = Task(
			function=function, task_id=task_id, args=args, kwargs=kwargs, time_unit=self._time_unit,
			cpu_count=cpu_count
		)
		return task

	def _add_task_to_to_do(self, task):
		self.to_do_queue.append(task)
		self._tasks_status[task.id] = 'added'
		self._task_counter += 1

	def add_task(self, function, args=None, kwargs=None, task_id=None, cpu_count=1):
		task = self._create_task(function=function, args=args, kwargs=kwargs, task_id=task_id, cpu_count=cpu_count)
		self._add_task_to_to_do(task=task)
		return task.id

	def get_full_outcome(self, task_id):
		return self.processed[task_id]

	def get_result(self, task_id):
		return self.get_full_outcome(task_id=task_id).result

	@property
	def incomplete_task_ids(self):
		"""
		tasks that are in worker_doing but the worker stops before they finish
		:rtype: set
		"""
		return set(self._incomplete_task_ids.keys())

	def process_done_queue(self, echo=0):
		count = 0
		processed = set()
		while True:
			try:
				outcome = self._get_from_done_queue()
			except (queue.Empty, IndexError):
				break
			else:
				self._tasks_status[outcome.task_id] = 'processed'
				self.processed[outcome.task_id] = outcome
				processed.add(outcome.task_id)
				count += 1
		if echo:
			print(f'{count} tasks processed')
		return processed

	def get_tasks_timing_summary(self):
		"""
		:rtype: DataFrame
		"""
		self.process_done_queue()
		return DataFrame.from_records([
			result.timestamp_record
			for result in self.processed.values()
		])

	@property
	def task_signature_data(self):
		"""
		:rtype: DataFrame
		"""
		self.process_done_queue()
		return DataFrame.from_records([
			result.signature
			for result in self.processed
		])

	def _generate_worker_id(self, prefix='controller'):
		self._worker_id_counter += 1
		return f'{prefix}_{self._worker_id_counter}'

	def get_worker_reports_summary(self, exclude_empty_reports=True):
		result = DataFrame.from_records([
			report.record
			for report in self.worker_reports.values()
		])
		if result.shape[0] > 0:
			if exclude_empty_reports:
				result = result[result['task_count'] > 0]
			result = result.sort_values('start_time').reset_index(drop=True)
		return result

	def do(self, *args, **kwargs):
		raise NotImplementedError('do is not implemented for BaseController')



from pandas import DataFrame

def join_and_preserve_types(left, right, priority='left', **kwargs):
	"""
	:type left: DataFrame
	:type right: DataFrame
	:type priority: str
	:type kwargs: dict
	:rtype: DataFrame
	"""

	if priority=='left':
		orig = left.dtypes.to_dict()
		orig.update(right.dtypes.to_dict())
	else:
		orig = right.dtypes.to_dict()
		orig.update(left.dtypes.to_dict())

	joined = left.merge(right=right, **kwargs)

	def _fix_type(x):
		try:
			return x.astype(orig[x.name])
		except:
			return x

	return joined.apply(lambda x: _fix_type(x))


from atlantis.ds.parallel_computing._Task import Task
import traceback
from pandas import DataFrame
import shap
from xgboost import XGBRegressor, XGBClassifier
from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
import matplotlib.pyplot as plt
from ...feature_importance import get_feature_importances
from .._get_data_from_namespace import get_obj_from_namespace, get_data_from_namespace


class LearningTask(Task):
	def __init__(
			self, project_name, estimator_class, estimator_name, estimator_id, estimator_arguments,
			training_test_slice_id, y_column, x_columns, evaluation_function,
	):

		super().__init__(project_name=project_name, task_id=None)
		if not isinstance(estimator_id, (str, int)):
			raise TypeError('estimator_id should be an int or str')

		if not isinstance(project_name, (str, int)):
			raise TypeError('project_name should be either an int or a str')

		if not isinstance(estimator_class, type):
			raise TypeError('estimator_class should be a type')

		if not isinstance(estimator_arguments, dict):
			raise TypeError('kwargs should be an int')

		if not isinstance(y_column, str):
			raise TypeError('y_column should be a str')

		self._estimator_id = estimator_id
		self._estimator_class = estimator_class
		self._estimator_name = estimator_name
		self._estimator_arguments = estimator_arguments

		self._training_test_id = training_test_slice_id
		self._y_column = y_column
		self._x_columns = x_columns
		self._evaluation_function = evaluation_function

		self._evaluation = None
		self._id = self.project_name, self.estimator_name, self.estimator_id, self.training_test_id, self.y_column
		self._predictions = None
		self._trained_estimator = None
		self._feature_importances = None
		self._shap_values = None
		self._training_x = None

	@property
	def training_test_id(self):
		return self._training_test_id

	@property
	def time_estimate_id(self):
		return self.estimator_name

	@property
	def status(self):
		return self._status

	def __hash__(self):
		return hash(self.id)

	@property
	def estimator_id(self):
		return self._estimator_id

	@property
	def estimator_class(self):
		return self._estimator_class

	@property
	def estimator_name(self):
		return self._estimator_name

	@property
	def estimator_arguments(self):
		return self._estimator_arguments

	@property
	def y_column(self):
		return self._y_column

	@property
	def x_columns(self):
		return self._x_columns

	@property
	def evaluation(self):
		return self._evaluation

	def evaluate(self, actual, predicted):
		self._evaluation = self._evaluation_function(actual=actual, predicted=predicted)

	@property
	def record(self):
		"""
		:rtype: dict
		"""
		evaluation = self.evaluation or {}

		return {
			'project_name': self.project_name,
			'estimator_name': self.estimator_name,
			'estimator_id': self.estimator_id,
			'training_test_id': self._training_test_id,
			'worker_id': self._worker_id,
			'status': self._status,
			'starting_time': self.starting_time,
			'ending_time': self.ending_time,
			'elapsed_ms': self.get_elapsed(unit='ms'),
			**evaluation
		}

	@property
	def predictions(self):
		"""
		:rtype: DataFrame
		"""
		if self._predictions is None:
			raise RuntimeError('predictions not available. you should run do() with return_predictions=True')
		return self._predictions

	@property
	def trained_estimator(self):
		"""
		:rtype: LinearRegression
		"""
		if self._trained_estimator is None:
			raise RuntimeError('trained estimator not available. you should run do() with return_predictions=True')
		return self._trained_estimator

	@property
	def feature_importances(self):
		"""
		:rtype: dict
		"""
		return self._feature_importances

	def do(self, namespace, worker_id, return_predictions=False):
		"""
		:type namespace: Namespace
		:type worker_id: int or str
		"""
		try:
			self.start()

			estimator = self.estimator_class(**self.estimator_arguments)
			training_test_slice = get_obj_from_namespace(
				namespace=namespace,
				obj_type='tts', obj_id=self.training_test_id
			)
			data = get_data_from_namespace(namespace=namespace, data_id=training_test_slice.data_id)

			training_data = training_test_slice.get_training_data(data=data)
			training_data = training_data[training_data[self.y_column].notna()]

			test_data = training_test_slice.get_test_data(data=data)

			training_x = training_data[self.x_columns]
			training_y = training_data[self.y_column]
			estimator.fit(X=training_x, y=training_y)

			test_x = test_data[self.x_columns]
			actual_all = test_data[self.y_column]
			actual_evaluation = test_data[test_data[self.y_column].notna()][self.y_column]
			predicted_all = estimator.predict(test_x)
			predicted_evaluation = predicted_all[test_data[self.y_column].notna()]

			if return_predictions:
				result = test_x
				result['actual'] = actual_all
				result['predicted'] = predicted_all
				self._trained_estimator = estimator
				self._predictions = result
				self._feature_importances = get_feature_importances(model=estimator, columns=self.x_columns)
				if isinstance(
					self.trained_estimator,
					(
						XGBRegressor, XGBClassifier,
						RandomForestClassifier, RandomForestRegressor,
						DecisionTreeRegressor, DecisionTreeClassifier
					)
				):
					explainer = shap.TreeExplainer(self.trained_estimator)
					self._shap_values = explainer.shap_values(training_x)
					self._training_x = training_x

			self.evaluate(actual=actual_evaluation, predicted=predicted_evaluation)
			if self._evaluation is None:
				raise RuntimeError('evaluation is None')

			self.end(worker_id=worker_id)
		except Exception as error:
			self.add_error(error=error, trace=traceback.format_exc())

	def plot_shap(self, plot_type='bar', show=True, path=None, width=16, height=12):
		if show:
			shap.summary_plot(self._shap_values, self._training_x, show=show, plot_type=plot_type)
		else:
			figure = plt.gcf()
			if width is not None and height is not None:
				figure.set_size_inches(width, height)
			shap.summary_plot(self._shap_values, self._training_x, show=False, plot_type=plot_type)
			plt.savefig(path, dpi=300, bbox_inches='tight')



def _get_product(l1, l2):
	"""
	crosses two lists and returns tuples of form (x, y) for all x in l1 and all y in l2
	:rtype: list[tuple]
	"""
	return [(x, y) for x in l1 for y in l2]


def get_product(*args):
	"""
	crosses n lists and returns tuples of length n where n is the number of lists where
	each tuple is (x1, x2, ...xn) and xi is any element of list i.
	:rtype: list[tuple]
	"""
	if len(args) == 2:
		return _get_product(args[0], args[1])
	elif len(args) > 2:
		result = get_product(_get_product(args[0], args[1]), *args[2:])
		return [(x[0][0], x[0][1], *x[1:]) for x in result]
	elif len(args) == 1 and isinstance(args[0], (list, tuple)):
		return get_product(*args[0])



def multiply_dictionary(dictionary, key, values):
	"""
	:type dictionary: dict
	:type key: str
	:type values: list
	:rtype: list[dict]
	"""
	result = []
	if not isinstance(values, (list, tuple)):
		values = [values]

	for value in values:
		new_d = dictionary.copy()
		new_d[key] = value
		result.append(new_d)
	return result


def create_grid(dictionary):
	"""
	creates a list of dictionaries with all combinations of key-values
	create_grid({'a': [1,2,3], 'b': [10, 20]}) produces:
	[{'a': 1, 'b': 10}, {'a': 1, 'b': 20}, {'a': 2, 'b': 10}, {'a': 2, 'b': 20}, {'a': 3, 'b': 10}, {'a': 3, 'b': 20}]
	:param dict[str, list] dictionary: a dictionary of lists of parameter values
	:rtype: list[dict[str,]]
	"""
	result = [{}]
	for key, values in dictionary.items():
		new_result = []
		for dictionary in result:
			new_result += multiply_dictionary(dictionary=dictionary, key=key, values=values)
		result = new_result
	return result



def is_word(s):
	if not isinstance(s, str):
		raise TypeError(f'{s} if of type {type(s)}')

	return any([character.isalpha() for character in s])


def count_words(text):
	if not isinstance(text, str):
		raise TypeError(f'{text} if of type {type(text)}')

	return len([s for s in text.split() if is_word(s)])



from pandas import DataFrame, concat
import numpy as np


def get_display_function():
	try:
		from IPython.core.display import display
		return display
	except ImportError:
		return False


class Score:
	def __init__(self, main_metric):
		self._main_metric = main_metric
		self._score_dictionary = None
		self._score = None

	@property
	def score(self):
		return self._score

	@score.setter
	def score(self, score):
		if self._score is not None:
			raise RuntimeError('cannot overwrite score!')
		self._score = score

	@property
	def score_dictionary(self):
		return self._score_dictionary

	@score_dictionary.setter
	def score_dictionary(self, score_dictionary):
		if self._score_dictionary is not None:
			raise RuntimeError('cannot overwrite score dictionary!')
		self._score_dictionary = score_dictionary
		self._score = score_dictionary[self._main_metric]


class Scoreboard:
	def __init__(self, main_metric=None, lowest_is_best=True, best_score=0):
		self._estimators = set()
		self._training_test_ids = set()

		if not lowest_is_best and best_score == 0:
			raise ValueError(f'best score of 0 does not work when the highest score is the best!')

		self._best_score = best_score

		self._all_combinations = set()
		self._measured = {}
		self._unmeasured = {}

		self._lowest_is_best = lowest_is_best
		self._main_metric = main_metric

		self._measured_data = None
		self._mean_score_per_data = None
		self._mean_score_per_estimator = None
		self._best_possible_score_per_estimator = None

	@property
	def lowest_is_best(self):
		return self._lowest_is_best

	def _add_estimator_data_combination(self, estimator_name, estimator_id, training_test_id):
		key = estimator_name, estimator_id, training_test_id
		if key not in self._all_combinations:
			self._all_combinations.add(key)
			self._unmeasured[key] = Score(main_metric=self._main_metric)

	def add_estimator(self, estimator_name, estimator_id):
		if not isinstance(estimator_name, str):
			raise TypeError(f'estimator_name should be str but it is of type {type(estimator_name)}')

		key = estimator_name, estimator_id
		if key in self._estimators:
			raise KeyError(f'estimator {key} already exists!')
		self._estimators.add(key)
		for training_test_id in self.training_test_ids:
			self._add_estimator_data_combination(
				estimator_name=estimator_name, estimator_id=estimator_id, training_test_id=training_test_id
			)

	def add_training_test_id(self, training_test_id):
		if training_test_id in self._training_test_ids:
			raise KeyError(f'training_test_id: {training_test_id} already exists!')
		self._training_test_ids.add(training_test_id)
		for estimator_name, estimator_id in self.estimators:
			self._add_estimator_data_combination(
				estimator_name=estimator_name, estimator_id=estimator_id, training_test_id=training_test_id
			)

	@property
	def training_test_ids(self):
		"""
		:rtype: set
		"""
		return self._training_test_ids

	@property
	def estimators(self):
		"""
		:rtype: set[(str, int)]
		"""
		return self._estimators

	def make_stale(self):
		self._measured_data = None
		self._mean_score_per_data = None
		self._mean_score_per_estimator = None
		self._best_possible_score_per_estimator = None

	def add_score(self, estimator_name, estimator_id, training_test_id, score_dictionary):
		if not isinstance(score_dictionary, dict):
			raise TypeError(f'score_dictionary should be a dict but it is of type {type(score_dictionary)}')
		try:
			score = self._unmeasured[(estimator_name, estimator_id, training_test_id)]
		except KeyError as e:
			display(self._unmeasured)
			raise e
		score.score_dictionary = score_dictionary

		self._measured[(estimator_name, estimator_id, training_test_id)] = score
		self.make_stale()

		del self._unmeasured[(estimator_name, estimator_id, training_test_id)]

	def add_task_score(self, task):
		"""
		:type task: LearningTask
		"""
		if task.status == 'done':
			self.add_score(
				estimator_name=task.estimator_name,
				estimator_id=task.estimator_id,
				training_test_id=task.training_test_id,
				score_dictionary=task.evaluation
			)
		else:
			raise RuntimeError(f'{task} is not done, it is {task.status}')

	def _get_measured_records(self, evaluation=False):
		if len(self.training_test_ids) == 0:
			raise RuntimeError('training_test_ids is empty')
		if len(self.estimators) == 0:
			raise RuntimeError('estimators is empty')

		records = []
		for key, score in self._measured.items():
			estimator_name, estimator_id, training_test_id = key
			record = {
				'estimator_name': estimator_name, 'estimator_id': estimator_id,
				'training_test_id': training_test_id,
				'score': score.score
			}
			if evaluation:
				record = {**record, **score.score_dictionary}
			records.append(record)
		return records

	@property
	def measured_data(self):
		"""
		:rtype: DataFrame
		"""

		if self._measured_data is None:
			data = DataFrame.from_records(self._get_measured_records())
			aggregate = data.groupby(['estimator_name', 'estimator_id']).agg(['count', 'mean', 'min', 'max', 'std'])
			aggregate.sort_values(by=('score', 'mean'), ascending=self.lowest_is_best, inplace=True)
			self._measured_data = aggregate
		return self._measured_data

	@property
	def evaluation_mean(self):
		records = self._get_measured_records(evaluation=True)
		data = DataFrame.from_records(records).groupby(['estimator_name', 'estimator_id']).mean().reset_index()
		return data

	def _get_all_records_fill_unmeasured_with_best(self):
		records = self._get_measured_records()
		for key in self._unmeasured.keys():
			estimator_name, estimator_id, training_test_id = key
			records.append({
				'estimator_name': estimator_name, 'estimator_id':
					estimator_id, 'training_test_id': training_test_id,
				'score': self._best_score
			})

		return records

	@property
	def mean_score_per_data(self):
		"""
		:rtype: DataFrame
		"""
		if self._mean_score_per_data is None:
			records = self._get_measured_records()
			all_data_sets = DataFrame({'training_test_id': list(self.training_test_ids)})
			if len(records) == 0:
				all_data_sets['score'] = None

			else:
				data = DataFrame.from_records(self._get_measured_records())
				data = data[['training_test_id', 'score']]
				aggregate = data.groupby('training_test_id').mean().reset_index()

				all_data_sets = all_data_sets.merge(aggregate, on='training_test_id', how='left')

			all_data_sets.sort_values(by='score', ascending=self.lowest_is_best, inplace=True, na_position='first')

			self._mean_score_per_data = all_data_sets
		return self._mean_score_per_data

	@property
	def mean_score_per_estimator(self):
		"""
		:rtype: DataFrame
		"""
		if self._mean_score_per_estimator is None:
			data = DataFrame.from_records(self._get_measured_records())
			data = data[['estimator_name', 'estimator_id', 'score']]
			self._mean_score_per_estimator = data.groupby(['estimator_name', 'estimator_id']).mean().reset_index()
			self._mean_score_per_estimator.sort_values(by='score', ascending=self.lowest_is_best, inplace=True)
		return self._mean_score_per_estimator

	@property
	def best_possible_score_per_estimator(self):
		"""
		:rtype: DataFrame
		"""
		if self._best_possible_score_per_estimator is None:
			data = DataFrame.from_records(self._get_all_records_fill_unmeasured_with_best())
			try:
				data = data[['estimator_name', 'estimator_id', 'score']]
			except:
				display(data)
				raise
			aggregate = data.groupby(['estimator_name', 'estimator_id']).mean().reset_index()
			aggregate.sort_values(by='score', ascending=self.lowest_is_best, inplace=True)

			self._best_possible_score_per_estimator = aggregate
		return self._best_possible_score_per_estimator

	def _repr_pretty_(self, p, cycle):
		if cycle:
			p.text(repr(self))
		else:
			self.display(p=p)

	def display(self, num_rows=None, p=None):
		display = get_display_function()
		measured_data = self.measured_data
		if num_rows is not None:
			measured_data = measured_data.head(num_rows)

		if display is False:
			print('scoreboard:')
			print(measured_data)
		else:
			print('scoreboard:')
			display(measured_data)


from ._CrossValidation import TrainingTest, TimeSeriesValidation, CrossValidation, Validation
from ._TrainingTestContainer import TrainingTestContainer
from ._ValidationContainer import ValidationContainer
from ._get_cross_validation import get_cross_validation
from ._get_training_test import get_training_test
from ._EstimatorRepository import EstimatorRepository
from ._EstimatorRace import EstimatorRace
from ._Scoreboard import Scoreboard



from time import sleep
import multiprocess

from ._WorkerReport import SupervisorReport
from ._do_task import do_task


def supervise(
		to_do_queue, done_queue, processed, incomplete_task_ids,
		workers, workers_doing, worker_reports, supervisor_id,
		cpu_usage, max_cpu_count,
		worker_id_counter, supervisor_alive, supervisor_reports,
		manager,
		sleep_time=0.1, worker_sleep_time=0.1, echo=0
):
	"""
	:type to_do_queue: multiprocessing.Queue[Task]
	:type done_queue: multiprocessing.Queue[TaskResult]
	:type processed: list[TaskResult]
	:type workers: dict[str, multiprocessing.Process]
	:type workers_doing: dict
	:type worker_reports: dict
	:type incomplete_task_ids: dict
	:type supervisor_id: str or int
	:type cpu_usage: multiprocessing.Value
	:type max_cpu_count: int
	:type worker_id_counter: int
	:type supervisor_alive: multiprocessing.Value
	:type supervisor_reports: dict
	:type sleep_time: float
	:type worker_sleep_time: float
	:type echo: bool or int
	"""
	report = SupervisorReport(supervisor_id=supervisor_id)
	report.log(f'supervisor {supervisor_id} started')

	worker_id_counter = 0

	while supervisor_alive.value == 1:

		# process done queue
		while True:
			try:
				result = done_queue.get_nowait()
			except:
				break
			else:
				processed.append(result)
				report.log(f'task {result.task_id} processed')

		# remove dead workers
		dead_worker_ids = set()
		if workers is not None:
			for worker_id, worker in workers.items():
				if not worker.is_alive():
					dead_worker_ids.add(worker_id)
					if worker in workers_doing:
						incomplete_task_id = workers_doing[worker_id]
						if incomplete_task_id is not None:
							incomplete_task_ids[incomplete_task_id] = 1
							report.log(f'incomplete task {incomplete_task_id} captured')
			for dead_worker_id in dead_worker_ids:
				del worker_id[dead_worker_id]
				report.log(f'dead worker {dead_worker_id} removed')

		# add workers as needed
		if workers is not None:
			while not to_do_queue.empty() and cpu_usage.value < max_cpu_count:

				# add worker
				new_worker_id = worker_id_counter.value
				worker_id_counter.value += 1
				process = multiprocess.Process(
					target=do_task,
					kwargs={
						'to_do_queue': to_do_queue,
						'done_queue': done_queue,
						'workers_doing': workers_doing,
						'worker_reports': worker_reports,
						'worker_id': new_worker_id,
						'cpu_usage': cpu_usage,
						'max_cpu_count': max_cpu_count,
						'sleep_time': worker_sleep_time,
						'echo': echo
					}
				)
				workers[new_worker_id] = process
				process.start()
				report.log(f'new worker {new_worker_id} added')
				sleep(sleep_time)

		sleep(sleep_time)

	report.log(f'supervisor {supervisor_id} ends')
	report.end()
	supervisor_reports.append(report)
	return True



class AtlanteanError(Exception):
	pass


class MissingArgumentError(AtlanteanError, ValueError):
	pass


class AmbiguousArgumentsError(AtlanteanError, ValueError):
	pass


class AmbiguityError(AtlanteanError):
	pass


class MutationError(AtlanteanError):
	pass


class FunctionNotImplementedError(AtlanteanError, AttributeError):
	pass


class ComparisonError(AtlanteanError):
	pass


class EmptyCollectionError(AtlanteanError, ValueError):
	pass



import re
from ..numbers import FlexibleNumber, NumberPart
from .get_month import get_month_num, get_month_name


class YearMonth(FlexibleNumber):

	def __init__(self, x=None, year=None, month=None, date=None, parts=None, sep='-', month_as='num'):
		# display_month can be 'num' 'abr' or 'ful'
		try:
			year = date.year
			month = date.month
		except:
			pass

		if type(x) is int:
			year = x // 100
			month = x % 100
		else:
			try:
				year = x.year
				month = x.month
			except:
				x = str(x)
				if re.match(pattern=r'^\d{4}.*\d{2}$', string=x):
					year = int(re.findall(pattern=r'^\d+', string=x)[0][:4])
					month = int(re.findall(pattern=r'\d+$', string=x)[0][-2:])
				elif re.match(pattern=r'^\d{4}\W*[a-zA-Z]+$', string=x):
					year = int(re.match(pattern=r'^\d+', string=x)[0][:4])
					month = get_month_num(name=re.search(pattern=r'[a-zA-Z]+$', string=x)[0])

		try:
			year_part = NumberPart(value=year, base=None, digits=4)
			month_part = NumberPart(value=month, base=12, start=1)
		except:
			year_part = parts[0]
			month_part = parts[1]

		if sep is None:
			sep = '-'

		self._month_as = month_as

		super().__init__(parts=[year_part, month_part], labels=['year', 'month'], sep=sep)
		self.adjust()

	@property
	def year(self):
		return self.get('year')

	@property
	def month(self):
		return self.get('month')

	@year.setter
	def year(self, year):
		self.set(value=int(year), label='year')

	@month.setter
	def month(self, month):
		self.set(value=int(month), label='month')

	def get_monthname(self, abr=False):
		return get_month_name(number=self.month.value, abr=abr)

	def get_year_monthname(self, abr=False):
		return str(self.year) + self._sep + self.get_monthname(abr=abr)

	def to_int(self):
		self.adjust()
		return self.year.value * 100 + self.month.value

	def to_months(self):
		return self.get_total(label='month')

	def to_years(self):
		return self.get_total(label='year')

	def __str__(self):
		if self._month_as == 'num':
			return super().__str__()
		elif self._month_as == 'abr':
			return self.get_year_monthname(abr=True)
		else:
			return self.get_year_monthname(abr=False)

	def __repr__(self):
		return f'{str(self.year)}-{str(self.month)}'

	def __hash__(self):
		return hash(self.to_int())

	# object comparisons
	def __eq__(self, other):
		"""
		:type other: YearMonth
		:rtype: bool
		"""
		return self.to_int() == other.to_int()

	def __lt__(self, other):
		return self.to_int() < other.to_int()

	def __le__(self, other):
		return self.to_int() <= other.to_int()

	def __ne__(self, other):
		return self.to_int() != other.to_int()

	def __gt__(self, other):
		return self.to_int() > other.to_int()

	def __ge__(self, other):
		return self.to_int() >= other.to_int()



def is_hashable(obj):
	try:
		hash(obj)
		return True
	except TypeError:
		return False


def make_hashable(obj):
	try:
		return make_hashable(obj.__hashkey__())
	except AttributeError:
		if isinstance(obj, (tuple, list)):
			return tuple((make_hashable(e) for e in obj))

		if isinstance(obj, dict):
			return tuple(sorted((k, make_hashable(v)) for k, v in obj.items()))

		if isinstance(obj, (set, frozenset)):
			return tuple(sorted(make_hashable(e) for e in obj))

		return obj



class ImputationError(Exception):
	pass


class ColumnError(ImputationError):
	pass


class NoColumnsError(ColumnError):
	pass


class MissingColumnError(ColumnError):
	pass


class ColumnTypeError(ColumnError):
	pass


class InputError(ImputationError):
	pass


class IncludeExcludeClashError(ColumnError):
	pass


class ImputerNotFittedError(ImputationError):
	pass


class MissingModelError(ImputationError):
	pass



from ._EstimatorRepository import EstimatorRepository
from ._TrainingTestContainer import TrainingTestContainer
from ._ValidationContainer import ValidationContainer


class EstimatorRace:
	def __init__(self, num_jobs=1):
		self._estimator_repository = None
		self._training_test_containers = []
		self._performances = {}
		self._num_jobs = num_jobs

	@property
	def estimator_repository(self):
		"""
		:rtype: EstimatorRepository or NoneType
		"""
		return self._estimator_repository

	def append_estimator_repository(self, estimator_repository):
		"""
		:type estimator_repository: EstimatorRepository
		"""
		if self._estimator_repository is None:
			self._estimator_repository = estimator_repository
		else:
			self._estimator_repository = self._estimator_repository + estimator_repository

	def append_data_container(self, data_container):
		"""
		:type data_container: TrainingTestContainer or ValidationContainer
		"""
		if isinstance(data_container, TrainingTestContainer):
			self._training_test_containers.append(data_container)
		elif isinstance(data_container, ValidationContainer):
			for training_test_container in data_container.folds:
				self._training_test_containers.append(training_test_container)

	def append_model(self, model, parameters):
		"""
		:type model: LogisticRegression or LinearRegression
		:type parameters: dict
		"""
		if self._estimator_repository is None:
			self._estimator_repository = EstimatorRepository(estimator=model, estimator_arguments=parameters)

		else:
			self.estimator_repository.append(estimator=model, estimator_arguments=parameters)




from .get_time import get_now
from .get_elapsed import get_elapsed
from datetime import timedelta


class Timer:
	def __init__(self, unit='ms', start_now=True):
		self._unit = unit
		self._start_time = None
		self._duration = None
		self._end_time = None

		self._last_start = None
		self._last_end = None

		if start_now:
			self.start()

	def __hashkey__(self):
		return (self.__class__.__name__, self.unit, self._start_time, self._duration, self._end_time)

	def start(self):
		"""
		:rtype: NoneType
		"""
		if self._start_time is not None:
			raise RuntimeError('Timer has already started!')
		self._duration = None
		self._last_end = self._end_time
		self._end_time = None
		self._start_time = get_now()

	def reset_timer(self):
		"""
		:rtype: NoneType
		"""
		self._last_start = self._start_time
		self._last_end = self._end_time
		self._start_time = None
		self._end_time = None

	@property
	def last_start(self):
		return self._last_start

	@property
	def last_end(self):
		return self._last_end

	def get_elapsed(self, end_time=None):
		"""
		:rtype: float or timedelta
		"""
		if end_time is None:
			end_time = get_now()
		now = end_time
		if self._start_time is not None:
			if self._duration is None:
				return get_elapsed(start=self._start_time, end=now, unit=self._unit)
			else:
				return get_elapsed(start=self._start_time, end=now, unit=self._unit) + self.duration
		else:
			raise RuntimeError('Timer is not activated yet!')

	def stop(self):
		"""
		:rtype: NoneType
		"""
		self._end_time = get_now()
		self._duration = self.get_elapsed(end_time=self._end_time)
		self.reset_timer()

	@property
	def duration(self):
		"""
		:rtype: float or timedelta
		"""
		return self._duration

	@property
	def unit(self):
		return self._unit
